"""
The module :py:mod:`endaq.ide.measurement` provides an abstract representation of measurement types for
easily retrieving specific data from ``.ide`` files. Several functions in :py:mod:`endaq.ide` accept
combinations of :py:class:`MeasurementType` constants for filtering datasets by sensor type.

Measurement types are represented by a set of singleton instances of :py:class:`MeasurementType`.

.. code:: python3

    import endaq.ide
    from endaq.ide.measurement import *

    doc = endaq.ide.get_doc("https://info.endaq.com/hubfs/data/surgical-instrument.ide")
    endaq.ide.get_channels(doc, ACCELERATION)
    endaq.ide.get_channels(doc, TEMPERATURE+PRESSURE)

Strings may also be used, either in combination with or instead of, the :py:class:`MeasurementType` instances.
The strings can be abbreviated, but no shorter than three characters.

.. code:: python3

    endaq.ide.get_channels(doc, "acceleration")
    endaq.ide.get_channels(doc, "accel")
    endaq.ide.get_channels(doc, ACCELERATION+"temp")

"""

__all__ = ['ANY', 'ACCELERATION', 'ALTITUDE', 'ANG_RATE', 'AUDIO', 'DIRECTION',
           'FREQUENCY', 'GENERIC', 'GYRO', 'HUMIDITY', 'LIGHT', 'LOCATION',
           'MAGNETIC', 'ORIENTATION', 'PRESSURE', 'ROTATION', 'SPEED',
           'TEMPERATURE', 'TIME', 'VOLTAGE',
           'get_measurement_type', 'filter_channels', 'get_channels']

from fnmatch import fnmatch
from shlex import shlex

from idelib.dataset import Dataset, Channel, SubChannel

# ============================================================================
#
# ============================================================================


class MeasurementType:
    """ Singleton/sentinel marker object for filtering channels by measurement
        type.
    """
    # TODO: Fix nomenclature. "Singleton" may not be the correct term; there
    #  are multiple instances of MeasurementType, but not multiple *duplicate*
    #  ones. This may be closer to the "Sentinel" pattern, but the class has
    #  more functionality than just marking.

    types = {}  # Maps all measurement type 'key' strings to objects.
    names = {}  # Maps display names to objects.
    verbose = False

    def __new__(cls, *keys, labels=None, doc=None):
        """ Create and return a new object, if one with a matching `key` does
            not already exist. If it does, the previous instance is returned.
            Called prior to `__init__()`.

            The first argument is taken as the display name of the
            measurement type. Its first 3 characters are used as the default
            key (the string actually used in filtering). Functions that use
            `MeasurementType` objects can also use those key strings.
            Additional non-keyword arguments are taken as alternate keys.

            :param labels: A list/tuple of substrings to match against
                `Channel` and `SubChannel` unit labels, used to identify
                the appropriate `NeasurementType` instance. The `name` is
                automatically included.
            :param doc: A docstring for the `MeasurementType` instance.
            :returns: `MeasurementType`
        """
        obj = None
        name = keys[0].strip()
        for key in keys:
            key = str(key).strip().lower()[:3]
            obj = cls.types.get(key, None)
            if not obj:
                if name not in cls.names:
                    obj = super().__new__(cls)
                    obj._name = name
                    obj._key = key
                    obj._keys = set()  # Alternative keys for this object
                    if labels is not None:
                        labels = (labels.strip().lower(),) if isinstance(labels, str) else labels
                        labels = tuple(labels) if labels else ()
                        if name not in labels:
                            labels += (name.lower(),)
                    obj._labels = labels
                    obj.__doc__ = doc or name
                    cls.types[key] = obj
                    cls.names[name] = obj
                else:
                    obj = cls.names[name]
                    cls.types[key] = obj
            obj._keys.add(key)
        return obj

    def __str__(self):
        return self._key

    def __repr__(self):
        if not self.verbose:
            return f"<MeasurementType: {self._name}>"
        keys = ', '.join([repr(k) for k in self._keys if k != self.key])
        if keys:
            return f"<MeasurementType: {self._name} {self.key!r} (alt: {keys})>"
        return f"<MeasurementType: {self._name} {self.key!r})>"

    def __eq__(self, other):
        return str(self) == str(other)

    def __hash__(self):
        # For interoperability with 'key' strings (as dict keys, etc.)
        # e.g., `mtype == "acc"`
        return hash(self._key)

    def __add__(self, other):
        # Concatenates as strings, so query-like sequences can be built.
        return f"{self} {other}"

    def __radd__(self, other):
        # Concatenates as strings, so query-like sequences can be built.
        return f"{other} {self}"

    def __sub__(self, other):
        # Concatenates as strings (with this one's negated)
        return f"{self} -{other}"

    def __rsub__(self, other):
        # Concatenates as strings (with this one's negated)
        return f"{other} -{self}"

    def __or__(self, other):
        # Same as __add__(), but convenient for those used to using bitwise OR to combine flags
        return self + other

    def __neg__(self):
        # Negates (appends a `-`) so query-like strings can be built.
        return f"-{self.key}"

    def __getitem__(self, *args, **kwargs):
        # For basic required string interoperability
        return self._key.__getitem__(*args, **kwargs)

    def upper(self):
        # For basic required string interoperability
        return self._key.upper()

    def lower(self):
        # For basic required string interoperability
        return self._key.lower()

    @property
    def name(self):
        return self._name

    @property
    def key(self):
        return self._key

    def match(self, channel):
        """ Does the given object match this measurement type?

            :param channel: The object to test. Can be an `idelib.dataset.Channel`,
                `idelib.dataset.SubChannel`, or a or a string of a measurement
                type name.
            :return: `True` if the given object uses this measurement type.
                In the case of `idelib.dataset.Channel`, `True` is returned
                if any of its subchannels match.
        """
        if getattr(channel, '_measurementType', channel) == self:
            return True

        if isinstance(channel, Channel):
            if not isinstance(channel, SubChannel):
                return any(self.match(c) for c in channel.children)
            else:
                mt = channel.units[0]
        elif isinstance(channel, str):
            mt = channel
        else:
            raise TypeError(f"Cannot compare measurement types with {channel} ({type(channel)})")

        mt = mt.lower()
        return any(label in mt or fnmatch(mt, label) for label in self._labels)

# ============================================================================
#
# ============================================================================


ANY = MeasurementType("Any/all", "*", "any", "all",
    doc="Marker object for matching any/all measurement types",
    labels=("*",))
""" Marker object for matching any/all measurement types """

ACCELERATION = MeasurementType("Acceleration", "acc", "g",
    doc="Marker object for filtering channels with acceleration data",
    labels=())
""" Marker object for filtering channels with acceleration data """

ORIENTATION = MeasurementType("Orientation", "imu", "qua", "gyr",
    doc="Marker object for filtering channels with rotation/orientation data",
    labels=("quaternion", "euler", "orientation"))
"Marker object for filtering channels with rotation/orientation data"

ROTATION = MeasurementType("Rotation", "rot", "ang",
    doc="Marker object for filtering channels with angular change rate data",
    labels=("rotation", "gyro"))
"Marker object for filtering channels with angular change rate data"

AUDIO = MeasurementType("Audio", "mic",
    doc="Marker object for filtering channels with sound level data",
    labels=("mic",))
"Marker object for filtering channels with sound level data"

LIGHT = MeasurementType("Light", "lux",
    doc="Marker object for filtering channels with light intensity data",
    labels=("lux", "uv"))
"Marker object for filtering channels with light intensity data"

PRESSURE = MeasurementType("Pressure",
    doc="Marker object for filtering channels with air pressure data",
    labels=())  # pressures
"Marker object for filtering channels with air pressure data"

TEMPERATURE = MeasurementType("Temperature",
    doc="Marker object for filtering channels with temperature data",
    labels=())  # temperature
"Marker object for filtering channels with temperature data"

HUMIDITY = MeasurementType("Relative Humidity", "hum",
    doc="Marker object for filtering channels with (relative) humidity data",
    labels=())  # Humidity
"Marker object for filtering channels with (relative) humidity data"

LOCATION = MeasurementType("Location", "pos", "gps",
    doc="Marker object for filtering channels with location data",
    labels=("pos",))  # GPS
"Marker object for filtering channels with location data"

SPEED = MeasurementType("Speed",
    doc="Marker object for filtering channels with rate-of-speed data",
    labels=("velocity",))  # GPS Ground Speed
"Marker object for filtering channels with rate-of-speed data"

TIME = MeasurementType("Time", "epo",
    doc="Marker object for filtering channels with time data",
    labels=("epoch",))  # GPS Epoch Time
"Marker object for filtering channels with time data"

# For potential future use
GENERIC = MeasurementType("Generic/Unspecified", "adc", "raw",
    labels=("adc", "raw"))
ALTITUDE = MeasurementType("Altitude",
    doc="Marker object for filtering channels with altitude data",
    labels=())
VOLTAGE = MeasurementType("Voltage",
    doc="Marker object for filtering channels with voltmeter data",
    labels=("volt",))
DIRECTION = MeasurementType("Direction",
    doc="Marker object for filtering channels with 2D directional data",
    labels=("compass", "heading"))
MAGNETIC = MeasurementType("Magnetic Field", "emf",
    doc="Marker object for filtering channels with magnetic field strength data",
    labels=("emf", "magnetic"))
FREQUENCY = MeasurementType("Frequency",
    doc="Marker object for filtering channels with frequency data",
    labels=("rate",))

# Synonyms, for convenience.
# TODO: Include abbreviations (e.g., TEMP = TEMPERATURE) for convenience? Current names are long.
ADC = GENERIC
ANG_RATE = ROTATION
GYRO = ROTATION
RAW = GENERIC
QUATERNION = ORIENTATION


# ============================================================================
#
# ============================================================================

def get_measurement_type(channel):
    """ Get the appropriate :py:class:`MeasurementType` object for a given
        `SubChannel`.
        
        Calling with a `Channel` returns a list of :py:class:`MeasurementType`
        objects, with one for each child `SubChannel`.

        :param channel: A `Channel` or `SubChannel` instance (e.g., from a
            `Dataset`).
        :returns: A :py:class:`MeasurementType` object (for a `SubChannel`), or
            a list of :py:class:`MeasurementType` objects (one for each child)
            if a `Channel` was supplied.
    """
    mt = getattr(channel, "_measurementType", None)
    if mt:
        return mt

    if channel.children:
        # Note: this caches the MeasurementType in the Channel/SubChannel; the
        # attribute `_measurementType` must be settable
        channel._measurementType = [get_measurement_type(c) for c in channel.children]
        return channel._measurementType
    else:
        for m in MeasurementType.types.values():
            if m == ANY:
                continue
            if m.match(channel):
                channel._measurementType = m
                return m

    return None


def split_types(query):
    """ Split a string of multiple :py:class:`MeasurementType` keys (e.g., one
        generated by using addition or subtraction of :py:class:`MeasurementType`
        objects and/or strings).

        :param query: A :py:class:`MeasurementType` or a string containing
            multiple :py:class:`MeasurementType` keys. A key can be excluded by
            prefixing it with a ``-``.
        :returns: A pair of sets of :py:class:`MeasurementType` instances: ones
            to include, and ones to exclude.
    """
    query = str(query).lower().strip()
    if query == "*":
        return set(MeasurementType.types.values()), set()

    inc = set()
    exc = set()
    prev = None

    # NOTE: Casting to string and parsing isn't always required, but this
    #   isn't used often enough to require high performance optimization.
    for token in shlex(query):
        token = token.lower()[:3]
        if token in MeasurementType.types:
            if prev == "-":
                exc.add(MeasurementType.types[token])
            else:
                inc.add(MeasurementType.types[token])
        elif token not in "+-":
            raise TypeError(f"Unknown measurement type: {token!r}")
        prev = token
    return inc, exc


def filter_channels(channels, measurement_type=ANY):
    """ Filter a list of `Channel` and/or `SubChannel` instances by their
        measurement type(s).

        :param channels: A list or dictionary of channels/subchannels to filter.
        :param measurement_type: A :py:class:`MeasurementType`, a measurement
            type 'key' string, or a string of multiple keys generated by adding
            and/or subtracting :py:class:`MeasurementType` objects. Any
            'subtracted' types will be excluded.
    """
    # Note: This is separated from `get_channels()` so it can be used
    # elsewhere (i.e., to retrieve channels by type from a `Recorder`)
    if isinstance(channels, dict):
        channels = channels.values()

    if measurement_type == ANY:
        return list(channels)

    # Note: if no `inc`, only `exc` is used
    inc, exc = split_types(measurement_type)

    result = []
    for ch in channels:
        thisType = get_measurement_type(ch)

        if isinstance(thisType, (list, set, tuple)):
            # `ch` is a Channel with SubChannels
            if exc.intersection(thisType):
                # One or more subchannels excluded; exclude channel.
                continue
            elif not inc or inc.intersection(thisType):
                result.append(ch)
                continue

        elif thisType not in exc:
            if not inc or thisType in inc:
                result.append(ch)

    return result


def get_channels(dataset, measurement_type=ANY, subchannels=True):
    """ Get a list of `Channel` or `SubChannel` instances from a `Dataset` by
        their measurement type(s).

        :param dataset: The `Dataset` from which to retrieve the list.
        :param measurement_type: A :py:class:`MeasurementType`, a measurement
            type 'key' string, or a string of multiple keys generated by adding
            and/or subtracting :py:class:`MeasurementType` objects. Any
            'subtracted' types will be excluded.
        :param subchannels: If `False`, get only `Channel` objects. If `True`,
            get only `SubChannel` objects.
        :returns: A list of matching `SubChannel` instances from the `Dataset`.
    """
    # This is really simple; it is a convenience for users, particularly
    # novices for whom the use of `getPlots()` may not be obvious.
    # TODO: Should there only be `filter_channels()`, and have it do `if isinstance(dataset, Dataset):`?
    #   The `subchannels` argument isn't applicable to non-Datasets, however, which would be weird.
    if subchannels:
        return filter_channels(dataset.getPlots(sort=False), measurement_type)
    return filter_channels(dataset.channels, measurement_type)



from __future__ import annotations

import typing  # for `SupportsIndex`, which is Python3.8+ only
from typing import Optional, Union
from collections.abc import Sequence
import warnings

import numpy as np
import pandas as pd
from scipy import signal
import plotly.graph_objects as go
import plotly.express as px

from endaq.plot import rolling_min_max_envelope
from endaq.calc import filters, integrate, utils, shock, psd


def shock_vibe_metrics(
        df: pd.DataFrame,
        tukey_percent: float = 0.1,
        highpass_cutoff: Optional[float] = None,
        accel_units: str = "gravity",
        disp_units: str = "in",
        freq_splits: typing.Union[np.ndarray, list, tuple] = (0, 65, 300, 1500, None),
        detrend: typing.Literal["start", "mean", "median", None] = "median",
        zero: typing.Literal["start", "mean", "median"] = "start",
        include_integration: bool = True,
        include_pseudo_velocity: bool = False,
        damp: float = 0.05,
        init_freq: float = 1.0,
        bins_per_octave: float = 12,
        include_resultant: bool = False,
        display_plots: bool = False,
) -> pd.DataFrame:
    """
    Compute the following shock and vibration metrics for a given time series dataframe
        - Peak Absolute Acceleration
        - RMS Acceleration
        - Peak Frequency
        - RMS Acceleration in Defined Frequency Ranges with `freq_splits`
        - Peak Absolute Velocity
        - RMS Velocity
        - Peak Absolute Displacement
        - RMS Displacement
        - Peak Pseudo Velocity & Corresponding Frequency
        
    :param df: the input dataframe with an index defining the time in seconds or datetime and units of `accel_units`
    :param tukey_percent: the portion of the time series to apply a Tukey window (a taper that forces beginning and end
        to 0), default is 0.1

        *  Note that the RMS metrics will only be computed on the portion of time that isn't tapered
    :param highpass_cutoff: the cutoff frequency of a preconditioning highpass
        filter; if None, no filter is applied. For shock events, it is recommended to set this to None (the default),
        but it is recommended for vibration.
    :param accel_units: the units to display acceleration as, default is `"gravity"` which will be shortened to 'g'
        in labels, the unit conversion is handled using :py:func:`~endaq.calc.utils.convert_units()`
    :param disp_units: the units to display displacement as and velocity (divided by seconds), default is `"in"`,
        the unit conversion is handled using :py:func:`~endaq.calc.utils.convert_units()`
    :param freq_splits: the boundaries of the frequency bins for the RMS calculations; must be strictly increasing, if
        `None` is given for the last value (the default) it will set this as the sampling rate
    :param detrend: the output quantity driven to zero prior to the calculations

        *  `None` does nothing
        *  `"start"` forces the first datapoint to 0,
        *  `"mean"` chooses ``-np.mean(output)``
        *  `"median"` (default) chooses ``-np.median(output)``
    :param zero: the output quantity driven to zero inside each integration call

        *  `"start"` (default) forces the first datapoint to 0,
        *  `"mean"` chooses ``-np.mean(output)``
        *  `"median"` (default) chooses ``-np.median(output)``
    :param include_integration: if `True`, include the calculations of velocity and displacement.  Defaults to `True`.
    :param include_pseudo_velocity: if `True`, include the more time-consuming calculation of pseudo velocity.
        Defaults to `False`.
    :param damp: the damping coefficient used in the shock response calculation `ζ`, related to the Q-factor by
        `ζ = 1/(2Q)`; defaults to 0.05
    :param init_freq: the initial frequency in the sequence for the shock response calculation; if `None`,
        use the frequency corresponding to the data's duration, default is 1.0 Hz
    :param bins_per_octave: the number of frequencies per octave for the shock response calculation
    :param include_resultant: add a resultant (root sum of the squares) for each metric,
        calculated from the other input dataframe columns
    :param display_plots: display plotly figures of the min/max envelope of acceleration, velocity, displacement and
        PVSS (default as False)
    :return: a dataframe containing all the metrics, one computed per column of the input dataframe

    Here is an example calculating and displaying these metrics for the bearing dataset discussed in our blog
    `Top 12 Vibration Metrics to Monitor & How to Calculate Them <https://blog.endaq.com/top-vibration-metrics-to-monitor-how-to-calculate-them>`_

    .. code:: python

        import endaq
        endaq.plot.utilities.set_theme('endaq_light')
        import pandas as pd
        import plotly.express as px

        # Get Acceleration Data
        accel = pd.read_csv('https://info.endaq.com/hubfs/Plots/bearing_data.csv', index_col=0)

        # Calculate Metrics
        metrics = endaq.calc.stats.shock_vibe_metrics(accel, include_resultant=False, freq_splits=[0, 65, 300, None])

        # Generate Figure with Bar Plots
        fig = px.bar(
            metrics,
            x="variable", y="value", color='variable',
            hover_data = ["units"],
            facet_col="calculation", facet_col_wrap=3)
        fig.update_yaxes(matches=None, visible=False)
        fig.update_xaxes(visible=False)
        fig.for_each_annotation(lambda a: a.update(text=a.text.split("=")[-1]))
        fig.show()

    .. plotly::
       :fig-vars: fig

        import endaq
        endaq.plot.utilities.set_theme('endaq_light')
        import pandas as pd
        import plotly.express as px

        # Get Acceleration Data
        accel = pd.read_csv('https://info.endaq.com/hubfs/Plots/bearing_data.csv', index_col=0)

        # Calculate Metrics
        metrics = endaq.calc.stats.shock_vibe_metrics(accel, include_resultant=False, freq_splits=[0, 65, 300, None])

        # Generate Figure with Bar Plots
        fig = px.bar(
            metrics,
            x="variable", y="value", color='variable',
            hover_data = ["units"],
            facet_col="calculation", facet_col_wrap=3)
        fig.update_yaxes(matches=None, visible=False)
        fig.update_xaxes(visible=False)
        fig.for_each_annotation(lambda a: a.update(text=a.text.split("=")[-1]))
        fig.show()

    """
    freq_splits = np.array(freq_splits)

    # Get unit conversion
    accel_2_disp = utils.convert_units(units_in=accel_units, units_out=disp_units + '/s^2')
    if accel_units == 'gravity':
        accel_units = 'g'

    # Remove Offset
    if detrend == "start":
        df -= df.iloc[0]
    elif detrend == "mean":
        df -= df.mean(axis="index")
    elif detrend == "median":
        df -= df.median(axis="index")
    elif detrend is not None:
        raise ValueError(f'kwarg detrend was {detrend}, must be one of [None, "start", "mean", "median"]')

    # Apply Filters & Window
    df = filters.butterworth(df, low_cutoff=highpass_cutoff, tukey_percent=tukey_percent)

    # Integrate
    data_list = [df]
    if include_integration:
        [accel, vel, disp] = integrate.integrals(
            df * accel_2_disp, n=2, tukey_percent=tukey_percent, highpass_cutoff=highpass_cutoff, zero=zero)
        data_list = [df, vel, disp]

    # Calculate Peak & RMS
    metrics = pd.DataFrame()
    rms_start = int(tukey_percent / 2 * df.shape[0])
    rms_end = df.shape[0] - rms_start
    for label, units, data in zip(
            ['Acceleration', 'Velocity', 'Displacement'],
            [accel_units, disp_units + '/s', disp_units],
            data_list):

        # Display Plots
        if display_plots:
            rolling_min_max_envelope(data, plot_as_bars=True, opacity=0.7).update_layout(yaxis_title_text=label).show()

        # Calculate Absolute Peak
        peak = pd.DataFrame(data.abs().max()).reset_index()
        peak.columns = ['variable', 'value']
        if include_resultant:
            peak = pd.concat([peak, pd.DataFrame({
                'variable': ['Resultant'],
                'value': [data.pow(2).sum(axis=1).pow(0.5).abs().max()]
            })])
        peak['calculation'] = 'Peak Absolute ' + label
        peak['units'] = units

        # Calculate RMS
        rms_stats = pd.DataFrame(data.iloc[rms_start:rms_end].pow(2).mean() ** 0.5).reset_index()
        rms_stats.columns = ['variable', 'value']
        if include_resultant:
            rms_stats = pd.concat([rms_stats, pd.DataFrame({
                'variable': ['Resultant'],
                'value': [np.sum(rms_stats.value ** 2) ** 0.5]
            })])
        rms_stats['calculation'] = f'RMS {label}'
        rms_stats['units'] = units

        # Add to Metrics
        metrics = pd.concat([metrics, peak, rms_stats])

    # Peak Frequency
    fs = 1 / utils.sample_spacing(df)
    parseval = psd.welch(df, scaling='parseval', bin_width=fs / len(df))
    peak = parseval.idxmax().reset_index()
    peak.columns = ['variable', 'value']
    peak['calculation'] = 'Peak Frequency'
    peak['units'] = 'Hz'
    metrics = pd.concat([metrics, peak])

    # Create Labels for Frequency Range Splits
    freq_splits = np.array(freq_splits)
    if freq_splits[-1] is None:
        freq_splits[-1] = int(np.floor(fs))
    freq_splits = freq_splits[freq_splits <= fs]
    labels = np.array(
        ['RMS from ' + str(freq_splits[i]) + ' to ' + str(freq_splits[i + 1]) for i in range(len(freq_splits) - 1)])

    # Calculate RMS in Defined Frequency Ranges
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', '.*empty.*')
        rms_psd_breaks = psd.to_jagged(parseval, freq_splits, agg='sum') ** 0.5
    rms_psd_breaks.index = pd.Series(labels, name='calculation')
    if include_resultant:
        rms_psd_breaks['Resultant'] = rms_psd_breaks.sum(axis=1)
    rms_psd_breaks = rms_psd_breaks.reset_index().melt(id_vars='calculation')
    rms_psd_breaks['units'] = accel_units
    metrics = pd.concat([metrics, rms_psd_breaks])

    # Optionally Calculate PVSS
    if include_pseudo_velocity:
        freqs = utils.logfreqs(df, bins_per_octave=bins_per_octave, init_freq=init_freq)
        pvss = shock.shock_spectrum(df * accel_2_disp, freqs=freqs, damp=damp, mode='pvss')
        if include_resultant:
            pvss['Resultant'] = shock.shock_spectrum(df * accel_2_disp, freqs=freqs, damp=damp, mode='pvss',
                                                     aggregate_axes=True)['Resultant']

        if display_plots:
            px.line(pvss, log_x=True, log_y=True).update_layout(yaxis_title_text='Pseudo Velocity',
                                                                xaxis_title_text='Natural Frequency (Hz)').show()

        # Add Peak Value
        pvss_stats = pd.DataFrame(pvss.max()).reset_index()
        pvss_stats.columns = ['variable', 'value']
        pvss_stats['calculation'] = 'Peak Pseudo Velocity'
        pvss_stats['units'] = disp_units + '/s'
        metrics = pd.concat([metrics, pvss_stats])

        # Add Peak Frequency
        pvss_stats = pd.DataFrame(pvss.idxmax()).reset_index()
        pvss_stats.columns = ['variable', 'value']
        pvss_stats['calculation'] = 'Peak PVSS Frequency'
        pvss_stats['units'] = 'Hz'
        metrics = pd.concat([metrics, pvss_stats])

    # Return Metrics
    return metrics


def find_peaks(
        df: pd.DataFrame,
        time_distance: float = 1.0,
        add_resultant: bool = False,
        threshold: float = None,
        threshold_reference: typing.Literal["rms", "peak"] = "peak",
        threshold_multiplier: float = 0.1,
        use_abs: bool = True,
        display_plots: bool = False,
) -> pd.DataFrame:
    """
    Find the peak events of a given time series using the maximum across all input columns
    
    :param df: the input dataframe with an index defining the time in seconds or datetime
    :param time_distance: the minimum time in seconds between events, default is 1.0
    :param add_resultant: add a resultant (root sum of the squares) to the time series prior to finding peaks,
        calculated from the other input dataframe columns
    :param threshold: if `None` (default) this is ignored, but if defined this value is passed as the minimum threshold
        to define a shock event
    :param threshold_reference: if the threshold isn't defined, calculate it from:

        *  `"peak"` (default) the overall peak
        *  `"rms"` the RMS value of the time series
    :param threshold_multiplier: if the threshold isn't defined, multiply this by the `threshold_reference`,
        suggestions are:
        *  `0.1` (default) when using peak, to get all events greater than 10% of the overall peak
        *  `4.0` when using RMS, a typical Gaussian signal has a kurtosis of 3.0
    :param use_abs: use the absolute value of the data to define peak events, default is True
    :param display_plots: display a plotly figure of the time series with the peak events plotted over it (default as
        False)
    :return: an array of index locations

    .. seealso::
        - `SciPy find_peaks function <https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.find_peaks.html>`_
          Documentation for the SciPy function used to identify the peak events.

    Here's an example implementation using a 60M dataset that loads the data, finds the peaks, and plots with the peak
    events identified all very quickly
    
    .. code:: python

        import endaq
        endaq.plot.utilities.set_theme()
        import plotly.graph_objects as go

        #Get Accel
        accel = endaq.ide.get_primary_sensor_data('https://info.endaq.com/hubfs/ford_f150.ide',measurement_type='accel',
            time_mode='datetime')

        #Filter
        accel = endaq.calc.filters.butterworth(accel,low_cutoff=2)

        #Get Peak Indexes
        indexes = endaq.calc.stats.find_peaks(
            accel, time_distance=2,
            threshold_reference="rms", threshold_multiplier=5.0)

        #Generate a Dataframe with Just the Peak Events
        df_peaks = accel.iloc[indexes]

        #Generate Shaded Bar Plot of All Data
        fig = endaq.plot.rolling_min_max_envelope(accel, plot_as_bars=True, opacity=0.7)

        #Add Peaks & Display
        fig.add_trace(
            go.Scatter(
                x=df_peaks.index,
                y=df_peaks.abs().max(axis=1).to_numpy(),
                mode='markers', name='Peak Events', marker_symbol='x', marker_color='white'
                )
            )
        fig.show()

    .. plotly::
       :fig-vars: fig

        import endaq
        endaq.plot.utilities.set_theme()
        import plotly.graph_objects as go

        #Get Accel
        accel = endaq.ide.get_primary_sensor_data('https://info.endaq.com/hubfs/ford_f150.ide',measurement_type='accel',
            time_mode='datetime')

        #Filter
        accel = endaq.calc.filters.butterworth(accel,low_cutoff=2)

        #Get Peak Indexes
        indexes = endaq.calc.stats.find_peaks(
            accel, time_distance=2,
            threshold_reference="rms", threshold_multiplier=5.0)

        #Generate a Dataframe with Just the Peak Events
        df_peaks = accel.iloc[indexes]

        #Generate Shaded Bar Plot of All Data
        fig = endaq.plot.rolling_min_max_envelope(accel, plot_as_bars=True, opacity=0.7)

        #Add Peaks & Display
        fig.add_trace(
            go.Scatter(
                x=df_peaks.index,
                y=df_peaks.abs().max(axis=1).to_numpy(),
                mode='markers', name='Peak Events', marker_symbol='x', marker_color='white'
                )
            )
        fig.show()
    """
    df_unmodified = df.copy()

    # Add Resultant
    if add_resultant:
        df['Resultant'] = df.pow(2).sum(axis=1).pow(0.5)

    # Reduce to Only Maximum Values Per Row
    if use_abs:
        peaks = df.abs().max(axis=1).to_numpy()
    else:
        peaks = df.max(axis=1).to_numpy()

    # Define Threshold
    if threshold is None:
        if threshold_reference == 'peak':
            threshold = np.max(peaks) * threshold_multiplier
        elif threshold_reference == 'rms':
            rms_val = df_unmodified.pow(2).mean() ** 0.5
            threshold = np.max(rms_val) * threshold_multiplier
            if add_resultant:
                threshold = np.sum(rms_val ** 2) ** 0.5 * threshold_multiplier

    # Find Peak Indexes
    indexes = signal.find_peaks(
        peaks,
        distance=int(time_distance / utils.sample_spacing(df)),
        height=threshold,
    )[0]

    # Optionally Display Plot
    if display_plots:
        df_peaks = df.iloc[indexes]
        fig = rolling_min_max_envelope(df, plot_as_bars=True, opacity=0.7)
        fig.add_trace(go.Scattergl(
            x=df_peaks.index,
            y=df_peaks.abs().max(axis=1).to_numpy(),
            mode='markers',
            name='Peak Events'
        ))
        fig.show()

    # Return Peak Times
    return indexes


def rolling_metrics(
        df: pd.DataFrame,
        indexes: np.array = None,
        index_values: np.array = None,
        num_slices: int = 100,
        slice_width: float = None,
        **kwargs,
) -> pd.DataFrame:
    """
    Quantify a series of time slices of a given time series
    
    :param df: the input dataframe with an index defining the time in seconds or datetime
    :param indexes: the index locations (not value) of each peak event to quantify like what is returned by :py:func:`~endaq.calc.stats.find_peaks()`
    :param index_values: the index values of each peak event to quantify (slower but more intuitive than using `indexes`)
    :param num_slices: the number of slices to split the time series into, default is 100,
        this is ignored if `indexes` or `index_values` are defined
    :param slice_width: the time in seconds to center about each slice index,
        if none is provided it will calculate one based upon the number of slices
    :param kwargs: Other parameters to pass directly to :py:func:`~endaq.calc.stats.shock_vibe_metrics()`
    :return: a dataframe containing all the metrics, one computed per column of the input dataframe, and one per peak event

    Here's a continuation of the example shown in :py:func:`~endaq.calc.stats.find_peaks()` that generates a table of
    metrics for a few defined time stamps, and then a row of subplots for each metric calculated.

    .. code:: python

        import endaq
        endaq.plot.utilities.set_theme()
        import plotly.express as px
        import pandas as pd

        # Get Accel
        accel = endaq.ide.get_primary_sensor_data('https://info.endaq.com/hubfs/ford_f150.ide',measurement_type='accel',
            time_mode='datetime')

        # Filter
        accel = endaq.calc.filters.butterworth(accel,low_cutoff=2)

        # Calculate for 3 Specific Times
        metrics = endaq.calc.stats.rolling_metrics(
            accel,
            index_values = pd.DatetimeIndex(['2020-03-13 23:40:13', '2020-03-13 23:45:00', '2020-03-13 23:50:00'],tz='UTC'),
            slice_width=5.0)

        # Simplify Timestamp Column
        metrics.timestamp = metrics.timestamp.astype(str).map(lambda x: x[10:19])

        # Generate Plot Table of Metrics
        table_plot = endaq.plot.table_plot(metrics)
        table_plot.show()

        # Calculate for 50 Equally Spaced & Sized Slices, Turning off Pseudo Velocity (Only Recommended for Smaller Time Slices)
        metrics = endaq.calc.stats.rolling_metrics(
            accel, num_slices=50, highpass_cutoff=2,
            tukey_percent=0.0, include_pseudo_velocity=False)


        # Generate Row with Subplots for each metric
        metrics_fig = px.scatter(
            metrics,
            x='timestamp',
            y='value',
            color='variable',
            facet_col='calculation',
            facet_col_spacing=0.03
        )
        metrics_fig.update_yaxes(title_text='', matches=None, showticklabels=True).update_xaxes(title_text='')
        metrics_fig.update_layout(width=3000, legend_y=1.2, legend_title_text='').for_each_annotation(lambda a: a.update(text=a.text.split("=")[-1]))
        metrics_fig.show()

    .. plotly::
       :fig-vars: table_plot, metrics_fig

        import endaq
        endaq.plot.utilities.set_theme()
        import plotly.express as px
        import pandas as pd

        # Get Accel
        accel = endaq.ide.get_primary_sensor_data('https://info.endaq.com/hubfs/ford_f150.ide',measurement_type='accel',
            time_mode='datetime')

        # Filter
        accel = endaq.calc.filters.butterworth(accel,low_cutoff=2)

        # Calculate for 3 Specific Times
        metrics = endaq.calc.stats.rolling_metrics(
            accel,
            index_values = pd.DatetimeIndex(['2020-03-13 23:40:13', '2020-03-13 23:45:00', '2020-03-13 23:50:00'],tz='UTC'),
            slice_width=5.0)

        # Simplify Timestamp Column
        metrics.timestamp = metrics.timestamp.astype(str).map(lambda x: x[10:19])

        # Generate Plot Table of Metrics
        table_plot = endaq.plot.table_plot(metrics)
        table_plot.show()

        # Calculate for 50 Equally Spaced & Sized Slices, Turning off Pseudo Velocity (Only Recommended for Smaller Time Slices)
        metrics = endaq.calc.stats.rolling_metrics(
            accel, num_slices=50, highpass_cutoff=2,
            tukey_percent=0.0, include_pseudo_velocity=False)


        # Generate Row with Subplots for each metric
        metrics_fig = px.scatter(
            metrics,
            x='timestamp',
            y='value',
            color='variable',
            facet_col='calculation',
            facet_col_spacing=0.03
        )
        metrics_fig.update_yaxes(title_text='', matches=None, showticklabels=True).update_xaxes(title_text='')
        metrics_fig.update_layout(width=3000, legend_y=1.2, legend_title_text='').for_each_annotation(lambda a: a.update(text=a.text.split("=")[-1]))
        metrics_fig.show()

    """
    indexes, slice_width, num, length = utils._rolling_slice_definitions(
        df,
        indexes=indexes,
        index_values=index_values,
        num_slices=num_slices,
        slice_width=slice_width
    )

    # Loop through and compute metrics
    metrics = pd.DataFrame()
    for i in indexes:
        window_start = max(0, i - num)
        window_end = min(length, i + num)
        event_metrics = shock_vibe_metrics(
            df.iloc[window_start:window_end],
            **kwargs
        )
        event_metrics['timestamp'] = df.index[i]
        metrics = pd.concat([metrics, event_metrics])

    return metrics


def L2_norm(
    array: np.ndarray,
    axis: Union[None, typing.SupportsIndex, Sequence[typing.SupportsIndex]] = None,
    keepdims: bool = False,
) -> np.ndarray:
    """
    Compute the L2 norm (a.k.a. the Euclidean Norm).

    :param array: the input array
    :param axis: the axis/axes along which to aggregate; if `None` (default),
        the L2 norm is computed along the flattened array
    :param keepdims: if `True`, the axes which are reduced are left in the
        result as dimensions of size one; if `False` (default), the reduced
        axes are removed
    :return: an array containing the computed values
    """
    return np.sqrt(np.sum(np.abs(array) ** 2, axis=axis, keepdims=keepdims))


def max_abs(
    array: np.ndarray,
    axis: Union[None, typing.SupportsIndex, Sequence[typing.SupportsIndex]] = None,
    keepdims: bool = False,
) -> np.ndarray:
    """
    Compute the maximum of the absolute value of an array.

    This function should be equivalent to, but generally use less memory than
    ``np.amax(np.abs(array))``.

    Specifically, it generates the absolute-value maximum from ``np.amax(array)``
    and ``-np.amin(array)``. Thus instead of allocating space for the
    intermediate array ``np.abs(array)``, it allocates for the axis-collapsed
    smaller arrays ``np.amax(array)`` & ``np.amin(array)``.

    .. note:: This method does **not** work on complex-valued arrays.

    :param array: the input data
    :param axis: the axis/axes along which to aggregate; if `None` (default),
        the absolute maximum is computed along the flattened array
    :param keepdims: if `True`, the axes which are reduced are left in the
        result as dimensions with size one; if `False` (default), the reduced
        axes are removed
    :return: an array containing the computed values
    """
    # Forbid complex-valued data
    if np.iscomplexobj(array):
        raise ValueError("`max_abs` does not accept complex arrays")

    return np.maximum(
        np.amax(array, initial=-np.inf, axis=axis, keepdims=keepdims),
        -np.amin(array, initial=np.inf, axis=axis, keepdims=keepdims),
    )


def rms(
    array: np.ndarray,
    axis: Union[None, typing.SupportsIndex, Sequence[typing.SupportsIndex]] = None,
    keepdims: bool = False,
) -> np.ndarray:
    """
    Calculate the root-mean-square (RMS) along a given axis.

    :param array: the input array
    :param axis: the axis/axes along which to aggregate; if `None` (default),
        the RMS is computed along the flattened array
    :param keepdims: if `True`, the axes which are reduced are left in the
        result as dimensions with size one; if `False` (default), the reduced
        axes are removed
    :return: an array containing the computed values
    """
    return np.sqrt(np.mean(np.abs(array) ** 2, axis=axis, keepdims=keepdims))


def rolling_rms(
    df: Union[pd.DataFrame, pd.Series], window_len: int, *args, **kwargs
) -> Union[pd.DataFrame, pd.Series]:
    """
    Calculate a rolling root-mean-square (RMS) over a pandas `DataFrame`.

    This function is equivalent to, but computationally faster than the following::

        df.rolling(window_len).apply(endaq.calc.stats.rms)

    :param df: the input data
    :param window_len: the length of the rolling window
    :param args: the positional arguments to pass into ``df.rolling().mean``
    :param kwargs: the keyword arguments to pass into ``df.rolling().mean``
    :return: the rolling-windowed RMS

    .. seealso::
        
        - `Pandas Rolling Mean <https://pandas.pydata.org/docs/reference/api/pandas.core.window.rolling.Rolling.mean.html>`_
          - official documentation for ``df.rolling().mean``
        - `Pandas Rolling Standard Deviation method <https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.core.window.rolling.Rolling.std.html>`_
          - similar to this function, but first removes the windowed mean before squaring
    """
    return df.pow(2).rolling(window_len).mean(*args, **kwargs).apply(np.sqrt, raw=True)



"""
Functions for retrieving summary data from a dataset.
"""
from __future__ import annotations
import typing

from collections import defaultdict
import datetime
import dateutil.tz
import warnings

import numpy as np
import pandas as pd
import pandas.io.formats.style
import idelib.dataset

from .measurement import MeasurementType, ANY, get_channels
from .files import get_doc
from .util import parse_time


__all__ = [
    "get_channel_table",
    "to_pandas",
    "get_primary_sensor_data",
]


# ============================================================================
# Display formatting functions
# ============================================================================

def format_channel_id(ch: idelib.dataset.Channel) -> str:
    """ Function for formatting an `idelib.dataset.Channel` or `SubChannel`
        for display. Renders as only the channel and subchannel IDs (the other
        information is shown in the rest of the table).

        :param ch: The `idelib.dataset.Channel` or `idelib.dataset.SubChannel`
            to format.
        :return: A formatted "channel.subchannel" string.
    """
    try:
        if ch.parent:
            return f"{ch.parent.id}.{ch.id}"
        else:
            return f"{ch.id}.*"

    except (AttributeError, TypeError, ValueError) as err:
        warnings.warn(f"format_channel_id({ch!r}) raised {type(err).__name__}: {err}")
        return str(ch)


def format_timedelta(val: typing.Union[int, float, datetime.datetime, datetime.timedelta]) -> str:
    """ Function for formatting microsecond timestamps (e.g., start, end,
        or duration) as times. Somewhat more condensed than the standard
        `DataFrame` formatting of `datetime.timedelta`.

        :param val: The `pandas.Timedelta` or `datetime.timedelta` to format.
            Will also work with microseconds as `float` or `int`.
        :return: A formatted time 'duration' string.
    """
    try:
        if isinstance(val, datetime.timedelta):
            td = pd.Timedelta(val)
        else:
            td = pd.Timedelta(microseconds=val)

        # NOTE: `components` attr only exists in pandas `Timedelta`
        c = td.components
        s = f"{c.minutes:02d}:{c.seconds:02d}.{c.milliseconds:04d}"
        if c.hours or c.days:
            s = f"{c.hours:02d}:{s}"
            if c.days:
                s = f"{c.days}d {s}"
        return s

    except (AttributeError, TypeError, ValueError) as err:
        warnings.warn(f"format_timedelta({val!r}) raised {type(err).__name__}: {err}")
        return str(val)


def format_timestamp(ts: typing.Union[int, float]) -> str:
    """ Function for formatting start/end timestamps. Somewhat more condensed
        than the standard Pandas formatting.

        :param ts: The timestamps in microseconds. Rendered as integers, since
            `idelib` timestamps have whole microsecond resolution.
        :return: A formatted timestamp string, with units.
    """
    try:
        return f"{int(ts)} µs"
    except (TypeError, ValueError) as err:
        warnings.warn(f"format_timestamp({ts!r}) raised {type(err).__name__}: {err}")
        return str(ts)

# ============================================================================
#
# ============================================================================


""" The default table formatting. """
TABLE_FORMAT = {
    'channel': format_channel_id,
    'start': format_timedelta,
    'end': format_timedelta,
    'duration': format_timedelta,
    'rate': "{:.2f} Hz",
}


def get_channel_table(dataset: typing.Union[idelib.dataset.Dataset, list],
                      measurement_type=ANY, 
                      start: typing.Union[int, float, str, datetime.datetime, datetime.timedelta] = 0,
                      end: typing.Optional[int, float, str, datetime.datetime, datetime.timedelta] = None,
                      formatting: typing.Optional[dict] = None,
                      index: bool = True, 
                      precision: int = 4,
                      timestamps: bool = False, 
                      **kwargs) -> typing.Union[pd.DataFrame, pd.io.formats.style.Styler]:
    """ Get summary data for all `SubChannel` objects in a `Dataset` that
        contain one or more type of sensor data. By using the optional
        `start` and `end` parameters, information can be retrieved for a
        specific interval of time.

        The `start` and `end` times, if used, may be specified in several
        ways:

        * `int`/`float` (Microseconds from the recording start)
        * `str` (formatted as a time from the recording start, e.g., `MM:SS`,
          `HH:MM:SS`, `DDd HH:MM:SS`). More examples:

          * ``":01"`` or ``":1"`` or ``"1s"`` (1 second)
          * ``"22:11"`` (22 minutes, 11 seconds)
          * ``"3:22:11"`` (3 hours, 22 minutes, 11 seconds)
          * ``"1d 3:22:11"`` (1 day, 3 hours, 22 minutes, 11 seconds)
        * `datetime.timedelta` or `pandas.Timedelta` (time from the
          recording start)
        * `datetime.datetime` (an explicit UTC time)

        :param dataset: A `idelib.dataset.Dataset` or a list of
            channels/subchannels from which to build the table.
        :param measurement_type: A :py:class:`~endaq.ide.MeasurementType`, a
            measurement type 'key' string, or a string of multiple keys
            generated by adding and/or subtracting
            :py:class:`~endaq.ide.MeasurementType` objects to filter the
            results. Any 'subtracted' types will be excluded.
        :param start: The starting time. Defaults to the start of the
            recording.
        :param end: The ending time. Defaults to the end of the recording.
        :param formatting: A dictionary of additional style/formatting items
            (see `pandas.DataFrame.style.format()`). If `False`, no additional
            formatting is applied.
        :param index: If `True`, show the index column on the left.
        :param precision: The default decimal precision to display. Can be
            changed later.
        :param timestamps: If `True`, show the start and end as raw
            microsecond timestamps.
        :returns: A table (`pandas.io.formats.style.Styler`) of summary data.
        :rtype: pandas.DataFrame
    """
    # We don't support multiple sessions on current Slam Stick/enDAQ recorders,
    # but in the event we ever do, this allows one to be specified like so:
    #       :param session: A `Session` or session ID to retrieve from a
    #           multi-session recording.
    # Leave out of docstring until we ever support it.
    session = kwargs.get('session', None)
    if session:
        session = getattr(session, 'sessionId', session)

    if hasattr(dataset, 'getPlots'):
        sources = get_channels(dataset, measurement_type)
    else:
        sources = dataset

    result = defaultdict(list)
    for source in sources:
        range_start = range_end = duration = rate = session_start = None
        samples = 0

        data = source.getSession(session)
        if data.session.utcStartTime:
            session_start = datetime.datetime.utcfromtimestamp(data.session.utcStartTime)
        start = parse_time(start, session_start)
        end = parse_time(end, session_start)

        if len(data):
            if not start and not end:
                start_idx, end_idx = 0, -1
                samples = len(data)
            else:
                start_idx, end_idx = data.getRangeIndices(start, end)
                end_idx = min(len(data) - 1, end_idx)
                if end_idx < 0:
                    samples = len(data) - start_idx - 1
                else:
                    samples = end_idx - start_idx

            range_start = data[int(start_idx)][0]
            range_end = data[int(end_idx)][0]
            duration = range_end - range_start
            rate = samples / (duration / 10 ** 6)

        result['channel'].append(source)
        result['name'].append(source.name)
        result['type'].append(source.units[0])
        result['units'].append(source.units[1])
        result['start'].append(range_start)
        result['end'].append(range_end)
        result['duration'].append(duration)
        result['samples'].append(samples)
        result['rate'].append(rate)

        # # TODO: RESTORE AFTER FIX IN idelib
        # dmin, dmean, dmax = data.getRangeMinMeanMax(start, end)
        # result['min'].append(dmin)
        # result['mean'].append(dmean)
        # result['max'].append(dmax)

    if formatting is False:
        return pd.DataFrame(result).style

    style = TABLE_FORMAT.copy()
    if timestamps:
        style.update({
            'start': format_timestamp,
            'end': format_timestamp
        })
    if isinstance(formatting, dict):
        style.update(formatting)

    styled = pd.DataFrame(result).style.format(style, precision=precision)
    if not index:
        return styled.hide_index()
    else:
        return styled

# ============================================================================
#
# ============================================================================

def get_utc_offset(dataset: idelib.dataset.Dataset) -> float:
    """
    Get a recorder's configured UTC time zone offset from an `idelib.Dataset`
    (i.e., an imported IDE file). Note that this is a user-configured option,
    and will be zero if the recorder did not have its UTC offset explicitly
    set.

    :param dataset: The IDE data from which to get the UTC offset.
    :return: The UTC offset, in seconds.
    """

    def crawl(parent):
        for el in parent:
            if el.name == 'ChannelDataBlock':
                return 0
            elif el.name in ('RecorderConfiguration', 'SSXBasicRecorderConfiguration'):
                return crawl(el)
            elif el.name == 'RecorderConfigurationList':
                for item in el:
                    data = item.dump()
                    if data.get('ConfigID') ==  0xBFF7F:
                        return data.get('IntValue', 0)
            elif el.name == 'UTCOffset':
                return el.value

        return 0

    return crawl(dataset.ebmldoc)


def to_pandas(
    channel: typing.Union[idelib.dataset.Channel, idelib.dataset.SubChannel],
    time_mode: typing.Literal["seconds", "timedelta", "datetime"] = "datetime",
    tz: typing.Union[pytz.timezone, dateutil.tz.tzfile, datetime.tzinfo,
                     typing.Literal["device", "local", "utc"]] = "utc"
) -> pd.DataFrame:
    """ Read IDE data into a pandas DataFrame.

        :param channel: a `Channel` object, as produced from `Dataset.channels`
            or :py:func:`endaq.ide.get_channels`
        :param time_mode: how to temporally index samples; each mode uses either
            relative times (with respect to the start of the recording) or
            absolute times (i.e., date-times):

            * `"seconds"` - a `pandas.Float64Index` of relative timestamps, in seconds
            * `"timedelta"` - a `pandas.TimeDeltaIndex` of relative timestamps
            * `"datetime"` - a `pandas.DateTimeIndex` of absolute timestamps

        :param tz: Optional time zone information for displaying the `"datetime"` time
            mode. It can be a standard time zone object (`pytz.timezone`,
            `dateutil.tz.tzfile`, `datetime.tzinfo`) or one of three special strings:

            * `"utc"` - standard UTC time (default).
            * `"local"` - the  current computer's local time zone (note: may not be the
                user's actual time zone when used on enDAQ Cloud).
            * `"device"` - the time zone specified by the original recording device's
                configured UTC offset.

        :return: a `pandas.DataFrame` containing the channel's data
    """
    time_mode = str(time_mode).casefold()
    if time_mode not in ('seconds', 'timedelta', 'datetime'):
        raise ValueError(f'invalid time mode {time_mode!r}')

    data = channel.getSession().arraySlice()
    t, data = data[0], data[1:].T
    t = (1e3*t).astype("timedelta64[ns]")

    if time_mode == "datetime":
        index = pd.to_datetime(t + np.datetime64(channel.dataset.lastUtcTime, "s"), utc=True)
        index.name = "timestamp"

        tz = tz.casefold() if isinstance(tz, str) else tz
        if tz != "utc":
            if tz == "device":
                tz = dateutil.tz.tzoffset('device', get_utc_offset(channel.dataset))
            elif tz == "local":
                tz = datetime.datetime.now().astimezone().tzinfo

            index = index.tz_convert(tz)
            
    else:
        if time_mode == "seconds":
            t = t / np.timedelta64(1, "s")

        index = pd.Series(t, name="timestamp")

    if hasattr(channel, "subchannels"):
        columns = [sch.name for sch in channel.subchannels]
    else:
        columns = [channel.name]

    return pd.DataFrame(data, index=index, columns=columns)

# ============================================================================
#
# ============================================================================


def get_primary_sensor_data(  
    name: str = "",
    doc: idelib.dataset.Dataset = None,
    measurement_type: typing.Union[str, MeasurementType] = ANY,
    criteria: typing.Literal["samples", "rate", "duration"] = "samples",
    time_mode: typing.Literal["seconds", "timedelta", "datetime"] = "datetime",
    tz: typing.Union[pytz.timezone, dateutil.tz.tzfile, datetime.tzinfo,
                     typing.Literal["device", "local", "utc"]] = "utc",
    least: bool = False,
    force_data_return: bool = False
) -> pd.DataFrame:
    """ Get the data from the primary sensor in a given .ide file using :py:func:`~endaq.ide.to_pandas()` 

        :param name: The file location to pull the data from, see :py:func:`~endaq.ide.get_doc()` 
            for more. This can be a local file location or a URL.
        :param doc: An open `Dataset` object, see :py:func:`~endaq.ide.get_doc()` 
            for more. If one is provided it will not attempt to use `name` to 
            load a new one.
        :param measurement_type: The sensor type to return data from, see :py:mod:`~endaq.ide.measurement`
            for more. The default is `"any"`, but to get the primary accelerometer
            for example, set this to `"accel"`.
        :param criteria: How to determine the "primary" sensor using the summary
            information provided by :py:func:`~endaq.ide.get_channel_table()`: 
        
            * `"sample"` - the number of samples, default behavior
            * `"rate"` - the sampling rate in Hz
            * `"duration"` - the duration from start to the end of data from that sensor
        :param time_mode: how to temporally index samples; each mode uses either
            relative times (with respect to the start of the recording) or
            absolute times (i.e., date-times):

            * `"seconds"` - a `pandas.Float64Index` of relative timestamps, in seconds
            * `"timedelta"` - a `pandas.TimeDeltaIndex` of relative timestamps
            * `"datetime"` - a `pandas.DateTimeIndex` of absolute timestamps
        :param tz: Optional time zone information for displaying the `"datetime"` time
            mode. It can be a standard time zone object (`pytz.timezone`,
            `dateutil.tz.tzfile`, `datetime.tzinfo`) or one of three special strings:

            * `"utc"` - standard UTC time (default).
            * `"local"` - the  current computer's local time zone (note: may not be the
                user's actual time zone when used on enDAQ Cloud).
            * `"device"` - the time zone specified by the original recording device's
                configured UTC offset.
        :param least: If set to `True` it will return the channels ranked lowest by
            the given criteria.
        :param force_data_return: If set to `True` and the specified `measurement_type`
            is not included in the file, it will return the data from any sensor 
            instead of raising an error which is the default behavior.

        :return: a `pandas.DataFrame` containing the sensor's data

        Here are some examples:
        
        .. code:: python3

            #Get sensor with the most samples, may return data of mixed units
            data = get_primary_sensor_data('https://info.endaq.com/hubfs/data/All-Channels.ide')

            #Instead get just the primary accelerometer data defined by number of samples
            accel = get_primary_sensor_data('https://info.endaq.com/hubfs/data/All-Channels.ide', measurement_type='accel')
    """
    
    #Get the doc object if it isn't provided
    if doc is None:
        doc = get_doc(name)
        
    #Get Channels of the Measurement Type
    channels = get_channel_table(doc, measurement_type).data
    
    #Raise error if measurement type isn't in the file
    if len(channels) == 0:
        error_str = f'measurement type "{measurement_type!r}" is not included in this file'
        if force_data_return:
            warnings.warn(error_str)
            channels = get_channel_table(doc, "any").data
        else:
            raise ValueError(error_str)
    
    #Filter based on criteria
    criteria = str(criteria).lower()
    if (criteria in ["samples", "rate", "duration"]):
        if least:
            channels = channels[channels[criteria] == channels[criteria].min()]
        else:
            channels = channels[channels[criteria] == channels[criteria].max()]
    else:        
        raise ValueError(f'invalid criteria "{criteria!r}"')
    
    #Get parent channel
    parent = channels.iloc[0].channel.parent
    
    #Get parent channel data
    data = to_pandas(parent, time_mode=time_mode, tz=tz)
    
    #Return only the subchannels with right units
    return data[channels.name]    



from collections import defaultdict, namedtuple
import warnings

from endaq.ide import to_pandas


class NoChannelException(Exception):
    pass


class ChannelStruct(namedtuple("ChannelStruct", "channel, sch_ids")):
    @property
    def eventarray(self):
        return self.channel.getSession()

    @property
    def fs(self):
        return self.eventarray.getSampleRate() if len(self.eventarray) > 0 else None

    @property
    def units(self):
        return self.channel[self.sch_ids[0]].units

    @property
    def axis_names(self):
        return [self.channel.subchannels[i].name for i in self.sch_ids]

    def to_pandas(self, *args, **kwargs):
        df = to_pandas(self.channel, *args, **kwargs).iloc[:, self.sch_ids]
        df.columns.name = "axis"
        return df


def chs_by_utype(dataset):
    """Group subchannels together by channel & utype."""
    for ch_id, channel in dataset.channels.items():
        # Group subchannel id's by unit type
        sch_ids_by_utype = defaultdict(list)
        for sch_id, subchannel in enumerate(channel.subchannels):
            sch_ids_by_utype[subchannel.units[0]].append(sch_id)

        # Separate each channel into subchannels of like-unit-type
        for utype, sch_ids in sch_ids_by_utype.items():
            yield utype, ChannelStruct(channel, sch_ids)


def map_utypes(iterable, utypes_map):
    """Re-label utype-schs pairs with new utypes."""
    return ((utypes_map.get(utype, utype), ch_struct) for utype, ch_struct in iterable)


def dict_groups(iterable):
    """Group an iterable of key-value pairs into a dict of lists of items."""
    result = defaultdict(list)
    for k, v in iterable:
        result[k].append(v)
    return result


def dict_chs_best(iterable, max_key=lambda x: len(x.eventarray)):
    """
    Group an iterable of utype-ch_struct pairs into a dict, keeping only those
    channels with the highest sample length for their utype.
    """
    result = {}
    for utype, ch_struct in iterable:
        if not utype in result:
            result[utype] = ch_struct
            continue

        result[utype] = max(result[utype], ch_struct, key=max_key)

    return result


def get_ch_type_best(dataset, utype, max_key=lambda x: len(x.eventarray)):
    """Get the highest sample-length acceleration channel from a recording."""
    chs = chs_by_utype(dataset)
    utype_chs = (ch_struct for (ut, ch_struct) in chs if ut == utype)

    try:
        return max(utype_chs, key=max_key)
    except ValueError:
        raise NoChannelException(f'no channels of type "{utype}" in recording')



"""
gdrive.py: Accessing data via Google Drive links.
"""
import re
from urllib.parse import parse_qs, urlparse

import requests

# DRIVE_URL = "https://docs.google.com/uc?export=download"
DRIVE_URL = "https://docs.google.com/uc"


def get_file_id(url):
    """
    Extract the Google Drive file ID from a URL. The ID can be in the
    URL path itself, or a query argument.

    :param url: The Google Drive 'shared link'.
    :return: The file ID portion of the URL.
    """
    # NOTE: This is a somewhat brittle hack. Revised later?
    parsed = urlparse(url)
    query = parse_qs(parsed.query)

    qid = query.get('id', '')
    if qid:
        return qid[0]

    match = re.match(r"^/file/d/(.*?)/view$", parsed.path)
    if match:
        return match.groups()[0]

    return None


def gdrive_download(url, localfile, params=None, cookies=None, drive_url=DRIVE_URL):
    """
    Retrieve an IDE from Google Drive. The file must be set to be shared
    with anyone with the URL.

    :param url: The 'shared link' to the file on Google Drive.
    :param localfile: The local filename (if saving the file).
    :param params: Additional (optional) request parameters.
    :param cookies: Optional browser cookies for the session.
    :param drive_url: The Google Docs download URL.
    :return: The 'get' response and the filename.
    """
    file_id = get_file_id(url)
    if not file_id:
        raise ValueError(f"Could not identify ID in URL {url}")

    p = {'id': file_id}
    if params:
        p.update(params)

    session = requests.Session()
    response = session.get(drive_url, params=p, cookies=cookies, stream=True)

    if not response.ok:
        raise ValueError(f"Could not retrieve data from URL {url} "
                         f"({response.status_code}: {response.reason})")

    # A response with the file will have its name in the headers.
    name = None

    if 'Content-Disposition' not in response.headers:
        # TODO: This is probably an intermediate page, possibly a warning.
        #  Get the download URL from the response body text and/or
        #  authorization token from the cookies, then get another response.
        raise ValueError(f"Could not retrieve data from URL {url} "
                         f"(not shared 'anyone with link'?)")

    if 'Content-Disposition' in response.headers:
        m = re.search('filename="(.*)"', ''.join(response.headers['Content-Disposition']))
        if m:
            name = m.groups()[0]

    if not name:
        raise ValueError(f"Could not retrieve data from URL {url}")

    return response, name



def cloud_correlation_script(files: list, file_download_url: str):
    """
    Creates a cloud dashboard (for use in cloud.endaq.com custom report generation) for comparing various recording
    files and exploring the correlations between them.  It creates a PCA plot with dropdown menus to select the
    PCA components, a plot which compares different file attribute values against eachother, a t-SNE plot, and
    all the attribute values.

    :param files: A `list` of JSON blobs of recording file data as it's given in a custom report in cloud.endaq.com
    :param file_download_url: The download URL for the IDE file most recently uploaded to cloud.endaq.com
    :return: The json based string which is to be given as the variable 'output' in enDAQ cloud
     custom dashboards (to produce the dashboard)
    """
    import numpy as np
    from endaq.plot.utilities import set_theme
    from endaq.cloud.core import json_table_to_df
    from endaq.plot import get_tsne_plot, get_pure_numpy_2d_pca, general_get_correlation_figure,\
        multi_file_plot_attributes
    from endaq.cloud import create_cloud_dashboard_output

    # Set the aesthetic theme of the figures, and store the list of colors used as a colorway for future use
    colorway = set_theme()['layout']['colorway']

    # Remove all file data which doesn't contain any attribute data
    files = [f for f in files if f['attributes']]

    # Ensure that there are at least 2 files
    if len(files) <= 2:
        raise ValueError(f"Must have at least 2 files given to analyze, was given {len(files)} instead")

    # Get the dataframe of the file attribute data
    df = json_table_to_df(files)

    # Create a dictionary mapping serial number of the enDAQ recording device with
    unique_serial_nums = np.unique(df['serial_number_id'])
    serials_to_colors = {serial_num: c for c, serial_num in
                        zip(colorway * (1 + len(unique_serial_nums) // len(colorway)), unique_serial_nums)}

    # The dictionary to be used to create the desired output of this function from (see the final line of this script
    # for how it's used to accomplish this)
    name_to_fig = dict()

    # Create PCA plot with 2 drop-down menus
    name_to_fig["PCA Plot"] = get_pure_numpy_2d_pca(df, color_discrete_map=serials_to_colors)

    # Create a plot that plots one attribute value against another attribute value, the two
    # attributes are selected via two dropdown menus
    name_to_fig["Attribute Correlation Explorer"] = general_get_correlation_figure(
        df,
        color_discrete_map=serials_to_colors,
        characteristics_to_show_on_hover=['id', 'recording_length', 'accelerationRMSFull', 'serial_number_id'],
    )

    # Create the t-SNE visualization
    name_to_fig["t-SNE"] = get_tsne_plot(df, color_discrete_map=serials_to_colors)

    # Generate Row of Subplots
    name_to_fig["Measurements Pane"] = multi_file_plot_attributes(
        df, recording_colors=df['serial_number_id'].map(serials_to_colors.get))

    # Create JSON Output
    output = create_cloud_dashboard_output(name_to_fig)

    return output



from __future__ import annotations

from dataclasses import dataclass
import typing
from typing import List, Tuple, Optional
import sys
import warnings

if sys.version_info[:2] >= (3, 8):
    from functools import cached_property
else:
    from backports.cached_property import cached_property

import numpy as np
import pandas as pd

import endaq.calc
from endaq.calc import psd, stats, shock, integrate, filters

from endaq.batch import quat
from endaq.batch.utils import ide_utils


MPS2_TO_G = 1 / 9.80665
MPS_TO_MMPS = 1000
MPS_TO_UMPS = 10 ** 6
MPS_TO_KMPH = 3600 / 1e3
M_TO_MM = 1000


@dataclass
class CalcParams:
    """
    The parameters for configuring the calculation routines in
    :py:class:`CalcCache`.

    Each of these parameters is *intentionally* left w/o a default value.
    Instead, defaults are provided at the function signatures for the
    :py:mod:`endaq.batch.core` functions. This ensures that data is passed
    correctly to them from the `CalcParam` object.
    """

    accel_highpass_cutoff: Optional[float]
    accel_integral_tukey_percent: float
    accel_integral_zero: typing.Literal["start", "mean", "median"]
    accel_start_time: Optional[np.timedelta64]
    accel_end_time: Optional[np.timedelta64]
    accel_start_margin: Optional[np.timedelta64]
    accel_end_margin: Optional[np.timedelta64]
    psd_freq_bin_width: float
    psd_window: str
    pvss_init_freq: float
    pvss_bins_per_octave: float
    vc_init_freq: float
    vc_bins_per_octave: float


class CalcCache:
    """
    A wrapper for `idelib.dataset.Dataset` that caches channel data streams as
    `pandas.Dataset`s.
    """

    PV_NATURAL_FREQS = np.logspace(0, 12, base=2, num=12 * 12 + 1, endpoint=True)

    def __init__(self, data, params: CalcParams):
        if (
            params.accel_start_time is not None
            and params.accel_start_margin is not None
        ):
            raise ValueError(
                "only one of `accel_start_time` and `accel_start_margin` may be set at once"
            )
        if params.accel_end_time is not None and params.accel_end_margin is not None:
            raise ValueError(
                "only one of `accel_end_time` and `accel_end_margin` may be set at once"
            )

        self._channels = data
        self._params = params

    @classmethod
    def from_ide(cls, dataset, params: CalcParams, preferred_chs: List[int] = []):
        """
        Instantiate a new `CalcCache` object from an IDE file.
        """
        # Rotation and Quaternion data should be grouped together
        utype_mapping = dict(
            Rotation="Rotation/Quaternion",
            Quaternion="Rotation/Quaternion",
        )

        data_valid = (
            (utype, ch_struct)
            for (utype, ch_struct) in ide_utils.chs_by_utype(dataset)
            if len(ch_struct.eventarray) > 0
        )
        data_grouped_best = ide_utils.dict_chs_best(
            (ide_utils.map_utypes(data_valid, utype_mapping)),
            max_key=lambda x: (x.channel.id in preferred_chs, len(x.eventarray)),
        )

        return cls(data_grouped_best, params=params)

    @dataclass
    class InputDataWrapper:
        data: pd.DataFrame
        units: Tuple[str, str]

        def to_pandas(self, time_mode="datetime"):
            expected_index_types = dict(
                timedelta=(pd.TimedeltaIndex, "TimedeltaIndex"),
                datetime=(pd.DatetimeIndex, "DatetimeIndex"),
                seconds=(
                    (pd.Float64Index, pd.Int64Index, pd.UInt64Index, pd.RangeIndex),
                    "{Float64/Int64/UInt64/Range}Index",
                ),
            )
            if not isinstance(self.data.index, expected_index_types[time_mode][0]):
                raise ValueError(
                    f"expected '{time_mode}' data index to be of type"
                    f"`{expected_index_types[time_mode][1]}`, "
                    f"instead found {type(self.data.index)}"
                )

            self.data.index.name = "timestamp"
            self.data.columns.name = "axis"

            return self.data

    @classmethod
    def from_raw_data(
        cls, data: List[Tuple[pd.DataFrame, Tuple[str, str]]], params: CalcParams
    ):
        """
        Instantiate a new `CalcCache` object from raw DataFrame / metadata pairs.
        """
        # Rotation and Quaternion data should be grouped together
        utype_mapping = dict(
            Rotation="Rotation/Quaternion",
            Quaternion="Rotation/Quaternion",
        )

        data_formatted = (
            (units[0], cls.InputDataWrapper(datum, units)) for (datum, units) in data
        )
        data_grouped = dict(ide_utils.map_utypes(data_formatted, utype_mapping))
        return cls(data_grouped, params=params)

    # ==========================================================================
    # Data Processing, just to make init cleaner
    # ==========================================================================

    @cached_property
    def _accelerationData(self):
        """Populate the _acceleration* fields, including splitting and extending data."""
        ch_struct = self._channels.get("Acceleration", None)
        if ch_struct is None:
            warnings.warn(f"no acceleration channel in data")
            return pd.DataFrame(
                np.empty((0, 3), dtype=float),
                index=pd.Series([], dtype="timedelta64[ns]", name="time"),
                columns=pd.Series(["X", "Y", "Z"], name="axis"),
            )

        aUnits = ch_struct.units[1]
        try:
            conversionFactor = {  # core units = m/s^2
                "g": 1 / MPS2_TO_G,
                "m/s\u00b2": 1,
            }[aUnits.lower()]
        except KeyError:
            raise ValueError(f'unknown acceleration channel units "{aUnits}"')

        aData = conversionFactor * ch_struct.to_pandas(time_mode="timedelta")
        dt = endaq.calc.sample_spacing(aData, convert=None)

        if self._params.accel_start_margin is not None:
            margin = int(np.ceil(self._params.accel_start_margin / dt))
            aData = aData.iloc[margin:]
        elif self._params.accel_start_time is not None:
            aData = aData.loc[self._params.accel_start_time :]
        if self._params.accel_end_margin is not None:
            margin = int(np.ceil(self._params.accel_end_margin / dt))
            aData = aData.iloc[: (-margin or None)]
        elif self._params.accel_end_time is not None:
            aData = aData.loc[: self._params.accel_end_time]

        aData = filters.butterworth(
            aData, low_cutoff=self._params.accel_highpass_cutoff
        )

        assert isinstance(aData, pd.DataFrame)
        return aData

    @cached_property
    def _accelerationResultantData(self):
        return pd.Series(
            stats.L2_norm(self._accelerationData.to_numpy(), axis=1),
            index=self._accelerationData.index,
        ).to_frame()

    @cached_property
    def _microphoneData(self):
        """Populate the _microphone* fields, including splitting and extending data."""
        ch_struct = self._channels.get("Audio", None)
        if ch_struct is None:
            return pd.DataFrame(
                np.empty((0, 1), dtype=float),
                index=pd.Series([], dtype="timedelta64[ns]", name="time"),
                columns=pd.Series(["mic"], name="axis"),
            )

        units = ch_struct.units[1]
        try:
            conversionFactor = {  # core units = Pa
                "pa": 1,
                "a": -5.307530522779073,
            }[units.lower()]
        except KeyError:
            raise ValueError(f'unknown microphone channel units "{units}"')

        return conversionFactor * ch_struct.to_pandas(time_mode="timedelta")

    @cached_property
    def _velocityData(self):
        aData = self._accelerationData
        if aData.size == 0:
            return aData

        if not self._params.accel_highpass_cutoff:
            warnings.warn(
                "no highpass filter used before integration; "
                "velocity calculation may be unstable"
            )

        return integrate._integrate(
            filters.butterworth(
                aData,
                low_cutoff=self._params.accel_highpass_cutoff,
                tukey_percent=self._params.accel_integral_tukey_percent,
            ),
            zero=self._params.accel_integral_zero,
        )

    @cached_property
    def _velocityResultantData(self):
        return pd.Series(
            stats.L2_norm(self._velocityData.to_numpy(), axis=1),
            index=self._velocityData.index,
        ).to_frame()

    @cached_property
    def _displacementData(self):
        vData = self._velocityData
        if vData.size == 0:
            return vData

        if not self._params.accel_highpass_cutoff:
            warnings.warn(
                "no highpass filter used before integration; "
                "displacement calculation may be unstable"
            )

        return integrate._integrate(
            filters.butterworth(
                vData,
                low_cutoff=self._params.accel_highpass_cutoff,
                tukey_percent=self._params.accel_integral_tukey_percent,
            ),
            zero=self._params.accel_integral_zero,
        )

    @cached_property
    def _displacementResultantData(self):
        return pd.Series(
            stats.L2_norm(self._displacementData.to_numpy(), axis=1),
            index=self._displacementData.index,
        ).to_frame()

    @cached_property
    def _PVSSData(self):
        aData = self._accelerationData
        if aData.size == 0:
            pvss = aData.copy()
            pvss.index.name = "frequency (Hz)"
            return pvss

        freqs = endaq.calc.logfreqs(
            aData, self._params.pvss_init_freq, self._params.pvss_bins_per_octave
        )
        freqs = freqs[
            (freqs >= 1 / (endaq.calc.sample_spacing(aData) * aData.shape[0]))
        ]
        pv = shock.shock_spectrum(
            aData,
            freqs,
            damp=0.05,
            mode="pvss",
        )

        return pv

    @cached_property
    def _PVSSResultantData(self):
        aData = self._accelerationData
        if aData.size == 0:
            pvss = self._accelerationResultantData.copy()
            pvss.index.name = "frequency (Hz)"
            return pvss

        freqs = endaq.calc.logfreqs(
            aData, self._params.pvss_init_freq, self._params.pvss_bins_per_octave
        )
        freqs = freqs[
            (freqs >= 1 / (endaq.calc.sample_spacing(aData) * aData.shape[0]))
        ]
        pv = shock.shock_spectrum(
            aData,
            freqs,
            damp=0.05,
            mode="pvss",
            aggregate_axes=True,
        )[['Resultant']]

        return pv

    @cached_property
    def _PSDData(self):
        aData = self._accelerationData
        if aData.size == 0:
            psdData = aData.copy()
            psdData.index.name = "frequency (Hz)"
            return psdData

        return endaq.calc.psd.welch(
            aData,
            bin_width=self._params.psd_freq_bin_width,
            window=self._params.psd_window,
            average="median",
        )

    @cached_property
    def _PSDResultantData(self):
        return self._PSDData.sum(axis="columns").to_frame()

    @cached_property
    def _VCCurveData(self):
        """Calculate Vibration Criteria (VC) Curves for the accelerometer."""
        psdData = self._PSDData
        if psdData.size == 0:
            vcData = psdData.copy()
            vcData.index.name = "frequency (Hz)"
            return vcData

        return psd.vc_curves(
            self._PSDData,
            fstart=self._params.vc_init_freq,
            octave_bins=self._params.vc_bins_per_octave,
        )

    @cached_property
    def _pressureData(self):
        """Populate the _pressure* fields, including splitting and extending data."""
        ch_struct = self._channels.get("Pressure", None)
        if ch_struct is None:
            return pd.DataFrame(
                np.empty((0, 1), dtype=float),
                index=pd.Series([], dtype="timedelta64[ns]", name="time"),
                columns=pd.Series(["Pressure"], name="axis"),
            )

        units = ch_struct.units[1]
        try:
            conversionFactor = {  # core units = kPa
                "pa": 1e-3,
                "psi": 6.89476,
                "atm": 101.325,
            }[units.lower()]
        except KeyError:
            raise ValueError(f'unknown pressure channel units "{units}"')

        return conversionFactor * ch_struct.to_pandas(time_mode="timedelta")

    @cached_property
    def _temperatureData(self):
        """Populate the _temperature* fields, including splitting and extending data."""
        ch_struct = self._channels.get("Temperature", None)
        if ch_struct is None:
            return pd.DataFrame(
                np.empty((0, 1), dtype=float),
                index=pd.Series([], dtype="timedelta64[ns]", name="time"),
                columns=pd.Series(["Temperature"], name="axis"),
            )

        units = ch_struct.units[1]
        try:
            conversionFactor, conversionOffset = {  # core units = degrees C
                "\xb0c": (1, 0),
                "\xb0k": (1, 273.15),
                "\xb0f": (5 / 9, -32 * (5 / 9)),
            }[units.lower()]
        except KeyError:
            raise ValueError(f'unknown temperature channel units "{units}"')

        return conversionOffset + conversionFactor * ch_struct.to_pandas(
            time_mode="timedelta"
        )

    @cached_property
    def _humidityData(self):
        """Populates the _humidity* fields, including splitting and extending data"""
        ch_struct = self._channels.get("Relative Humidity", None)
        if ch_struct is None:
            return pd.DataFrame(
                np.empty((0, 1), dtype=float),
                index=pd.Series([], dtype="timedelta64[ns]", name="time"),
                columns=pd.Series(["Relative Humidity"], name="axis"),
            )

        units = ch_struct.units[1]
        if units.lower() not in ("rh",):
            raise ValueError(f'unknown humidity channel units "{units}"')

        return ch_struct.to_pandas(time_mode="timedelta")

    @cached_property
    def _gyroscopeData(self):
        """Populate the _gyro* fields, including splitting and extending data."""
        ch_struct = self._channels.get("Rotation/Quaternion", None)
        if ch_struct is None:
            return pd.DataFrame(
                np.empty((0, 3), dtype=float),
                index=pd.Series([], dtype="timedelta64[ns]", name="time"),
                columns=pd.Series(["X", "Y", "Z"], name="axis"),
            )

        data = ch_struct.to_pandas(time_mode="timedelta")
        dt = endaq.calc.sample_spacing(data, convert="to_seconds")
        units = ch_struct.units[1]
        if units.lower() == "q":
            quat_raw = data[["W", "X", "Y", "Z"]].to_numpy()

            data = pd.DataFrame(
                (180 / np.pi) * quat.quat_to_angvel(quat_raw, dt, qaxis=1),
                index=data.index,
                columns=pd.Series(["X", "Y", "Z"], name="axis"),
            )

            def strip_invalid_prefix(data, prefix_len):
                """Search prefix for invalid data and remove it (if any)."""
                data_array = data.to_numpy()
                data_mag = stats.L2_norm(data_array[: 4 * prefix_len], axis=-1)
                # the derivative method for `quat_to_angvel` uses the *average*
                # of adjacent differences
                # -> any rotation spikes will result in two adjacent,
                # nearly-equal peaks
                data_agg = 0.5 * (data_mag[:-1] + data_mag[1:])
                argmax_prefix = np.argmax(data_agg[:prefix_len])

                # Prefix data is considered "anomalous" if it is much larger
                # than any surrounding data
                if data_agg[argmax_prefix] > 2 * data_mag[prefix_len:].max():
                    data = data.iloc[argmax_prefix + 2 :]

                return data

            data = strip_invalid_prefix(
                data, prefix_len=max(4, int(np.ceil(0.25 / dt)))
            )
        elif units.lower() not in ("dps", "deg/s"):
            raise ValueError(f'unknown gyroscope channel units "{units}"')

        return data

    @cached_property
    def _gpsPositionData(self):
        ch_struct = self._channels.get("Location", None)
        if ch_struct is None:
            return pd.DataFrame(
                np.empty((0, 2), dtype=float),
                index=pd.Series([], dtype="timedelta64[ns]", name="time"),
                columns=pd.Series(["Latitude", "Longitude"], name="axis"),
            )

        units = ch_struct.units[1]
        if units.lower() != "degrees":
            raise ValueError(f'unknown GPS position channel units "{units}"')

        # resampling destroys last values -> no resampling
        return ch_struct.to_pandas(time_mode="timedelta")

    @cached_property
    def _gpsSpeedData(self):
        ch_struct = self._channels.get("GNSS Speed", None)
        if ch_struct is None:
            return pd.DataFrame(
                np.empty((0, 1), dtype=float),
                index=pd.Series([], dtype="timedelta64[ns]", name="time"),
                columns=pd.Series(["GPS Speed"], name="axis"),
            )

        units = ch_struct.units[1]
        if units != "m/s":
            raise ValueError(f'unknown GPS ground speed channel units "{units}"')

        return MPS_TO_KMPH * ch_struct.to_pandas(time_mode="timedelta")

    # ==========================================================================
    # Analyses
    # ==========================================================================

    @cached_property
    def accRMSFull(self):
        """Accelerometer Tri-axial RMS."""
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            rms = self._accelerationData.apply(
                stats.rms, axis="rows", raw=True
            )  # RuntimeWarning: Mean of empty slice.

        rms["Resultant"] = stats.L2_norm(rms.to_numpy())
        rms.name = "RMS Acceleration"
        return MPS2_TO_G * rms

    @cached_property
    def velRMSFull(self):
        """Velocity Tri-axial RMS, after applying a 0.1Hz highpass filter."""
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            rms = self._velocityData.apply(
                stats.rms, axis="rows", raw=True
            )  # RuntimeWarning: Mean of empty slice.

        rms["Resultant"] = stats.L2_norm(rms.to_numpy())
        rms.name = "RMS Velocity"
        return MPS_TO_MMPS * rms

    @cached_property
    def disRMSFull(self):
        """Displacement Tri-axial RMS, after applying a 0.1Hz highpass filter."""
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            rms = self._displacementData.apply(
                stats.rms, axis="rows", raw=True
            )  # RuntimeWarning: Mean of empty slice.

        rms["Resultant"] = stats.L2_norm(rms.to_numpy())
        rms.name = "RMS Displacement"
        return M_TO_MM * rms

    @cached_property
    def accPeakFull(self):
        """Peak instantaneous tri-axial acceleration"""
        max_abs = self._accelerationData.abs().max(axis="rows")
        max_abs_res = self._accelerationResultantData.max(axis="rows")
        max_abs["Resultant"] = max_abs_res.iloc[0]
        max_abs.name = "Peak Absolute Acceleration"
        return MPS2_TO_G * max_abs

    @cached_property
    def pseudoVelPeakFull(self):
        """Peak Pseudo Velocity"""
        max_pv = self._PVSSData.max(axis="rows")
        max_pv_res = self._PVSSResultantData.max(axis="rows")
        max_pv["Resultant"] = max_pv_res.iloc[0]
        max_pv.name = "Peak Pseudo Velocity Shock Spectrum"
        return MPS_TO_MMPS * max_pv

    @cached_property
    def gpsLocFull(self):
        """Average GPS location"""
        data = self._gpsPositionData
        # 0's occur when gps doesn't have a "lock" -> remove them
        data = data.iloc[np.all(data.to_numpy() != 0, axis=1)]
        if data.size == 0:
            gpsPos = pd.Series(
                [np.nan, np.nan],
                index=pd.Series(["Latitude", "Longitude"], name="axis"),
            )
        else:
            gpsPos = data.iloc[-1]

        gpsPos.name = "GPS Position"
        return gpsPos

    @cached_property
    def gpsSpeedFull(self):
        """Average GPS speed"""
        data = self._gpsSpeedData
        # 0's occur when gps doesn't have a "lock" -> remove them
        data = data.iloc[data.to_numpy() != 0]

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            gpsSpeed = data.mean()  # RuntimeWarning: Mean of empty slice.

        gpsSpeed.name = "GPS Speed"
        return gpsSpeed

    @cached_property
    def gyroRMSFull(self):
        """Gyroscope RMS"""
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            rms = self._gyroscopeData.apply(
                stats.rms, axis="rows", raw=True
            )  # RuntimeWarning: Mean of empty slice.
        rms["Resultant"] = stats.L2_norm(rms.to_numpy())
        rms.name = "RMS Angular Velocity"
        return rms

    @cached_property
    def micRMSFull(self):
        """Microphone RMS"""
        warnings.warn("use `micDecibelsFull` instead", DeprecationWarning)

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            mic = self._microphoneData.apply(
                stats.rms, axis="rows", raw=True
            )  # RuntimeWarning: Mean of empty slice.

        mic.name = "RMS Sound Pressure"
        return mic

    @cached_property
    def micDecibelsFull(self):
        """Microphone dB"""
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            mic = self._microphoneData.apply(
                stats.rms, axis="rows", raw=True
            )  # RuntimeWarning: Mean of empty slice.
        mic = mic.apply(endaq.calc.to_dB, reference="SPL")

        mic.name = "Average Sound Pressure Level"
        return mic

    @cached_property
    def tempFull(self):
        """Average Temperature"""
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            temp = self._temperatureData.mean()  # RuntimeWarning: Mean of empty slice.

        temp.name = "Average Temperature"
        return temp

    @cached_property
    def pressFull(self):
        """Average Pressure"""
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            press = self._pressureData.mean()  # RuntimeWarning: Mean of empty slice.

        press.name = "Average Pressure"
        return press

    @cached_property
    def humidFull(self):
        """Average Relative Humidity"""
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            humid = self._humidityData.mean()  # RuntimeWarning: Mean of empty slice.

        humid.name = "Average Relative Humidity"
        return humid



import ebmlite


def iter_hierarchy(ebmldoc, nocache=False):
    """
    Iterate depth-first over all elements, yielding for each element its
    hierarchical path from the root document.

    :param nocache: whether to skip parser caching (defaults to False)
    :send: whether the current element's descendants should be skipped;
        defaults to None -> False
    :send type: bool
    :yield: the hierarchical path to an element from the document root for
        every element in the document
    :yield type: list
    """
    should_skip = yield [ebmldoc]
    if should_skip:
        return

    try:
        iter_subelements = ebmldoc.__iter__(nocache=nocache)
    except AttributeError:
        return

    for subelement in iter_subelements:
        iter_subhierarchy = iter_hierarchy(subelement, nocache=nocache)

        should_skip = None  # must start an iterator by sending `None`
        try:
            while True:
                should_skip = yield [ebmldoc] + iter_subhierarchy.send(should_skip)
        except StopIteration:
            continue


def filter_elements_by_name(doc, elem_names):
    """Filter hierarchical element paths by element names."""
    elem_names = [doc.name] + elem_names
    iter_elem_hierarchy = iter_hierarchy(doc)
    should_skip = None

    try:
        while True:
            element_path = iter_elem_hierarchy.send(should_skip)
            should_skip = element_path[-1].name != elem_names[len(element_path) - 1]
            if not should_skip and len(element_path) == len(elem_names):
                yield element_path
    except StopIteration:
        pass


def iter_config_attrs(filename, config_id, element_name):
    """Determine from file whether to perform segment processing."""
    with open(filename, "rb") as file:
        doc = ebmlite.loadSchema("mide_ide.xml").load(file)
        for elem_hierarchy in filter_elements_by_name(
            doc,
            [
                "RecorderConfigurationList",
                "RecorderConfigurationItem",
                "ConfigID",
            ],
        ):
            if elem_hierarchy[-1].value != config_id:
                continue
            for elem in elem_hierarchy[-2]:
                if elem.name == element_name:
                    yield elem.value



from __future__ import annotations

import typing
from typing import List, Optional, Iterable, Union, Callable
import functools

import numpy as np
import pandas as pd
import scipy.integrate

from endaq.calc import filters, utils


def _integrate(
    df: pd.DataFrame,
    zero: Union[typing.Literal["start", "mean", "median"], Callable] = "start",
) -> pd.DataFrame:
    """Integrate data over an axis."""
    dt = utils.sample_spacing(df)

    result = df.apply(
        functools.partial(scipy.integrate.cumulative_trapezoid, dx=dt, initial=0),
        axis=0,
        raw=True,
    )
    # In lieu of explicit initial offset, set integration bias to remove mean
    # -> avoids trend artifacts after successive integrations
    if zero == "start":
        return result

    if zero in ("mean", "median"):
        zero = dict(mean=np.mean, median=np.median)[zero]

    return result - result.apply(zero, axis="index", raw=True)


def iter_integrals(
    df: pd.DataFrame,
    zero: Union[typing.Literal["start", "mean", "median"], Callable] = "start",
    highpass_cutoff: Optional[float] = None,
    tukey_percent: float = 0.0,
) -> Iterable[pd.DataFrame]:
    """
    Iterate over conditioned integrals of the given original data.

    :param df: the input data
    :param zero: the output quantity driven to zero by the integration constant;
        `"start"` (default) chooses an integration constant of ``-output[0]``,
        `"mean"` chooses ``-np.mean(output)`` & `"median"` chooses
        ``-np.median(output)``
    :param highpass_cutoff: the cutoff frequency of a preconditioning highpass
        filter; if None, no filter is applied
    :param tukey_percent: the alpha parameter of a preconditioning Tukey filter;
        if 0 (default), no filter is applied
    :return: an iterable over the data's successive integrals; the first item
        is the preconditioned input data

    .. seealso::

        - `SciPy trapezoid integration <https://docs.scipy.org/doc/scipy/reference/generated/scipy.integrate.cumulative_trapezoid.html>`_
          Documentation for the integration function used internally.

        - `SciPy Butterworth filter design <https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.butter.html>`_
          Documentation for the butterworth filter design function used in
          preprocessing.

        - `SciPy Tukey window <https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.windows.tukey.html>`_
          Documentation for the Tukey window function used in preprocessing.
    """
    while True:
        yield df.copy()  # otherwise, edits to the yielded item would alter the results
        df = filters.butterworth(
            df,
            low_cutoff=highpass_cutoff,
            high_cutoff=None,
            tukey_percent=tukey_percent,
        )
        df = _integrate(df, zero)


def integrals(
    df: pd.DataFrame,
    n: int = 1,
    zero: Union[typing.Literal["start", "mean", "median"], Callable] = "start",
    highpass_cutoff: Optional[float] = None,
    tukey_percent: float = 0.0,
) -> List[pd.DataFrame]:
    """
    Calculate `n` integrations of the given data.

    :param df: the data to integrate, indexed with timestamps
    :param n: the number of integrals to calculate; defaults to `1`
    :param zero: the output quantity driven to zero by the integration constant;
        `"start"` (default) chooses an integration constant of ``-output[0]``,
        `"mean"` chooses ``-np.mean(output)`` & `"median"` chooses
        ``-np.median(output)``
    :param highpass_cutoff: the cutoff frequency for the initial highpass filter;
        this is used to remove artifacts caused by DC trends
    :param tukey_percent: the alpha parameter of a preconditioning Tukey filter;
        if 0 (default), no filter is applied
    :return: a length `n+1` list of the kth-order integrals from 0 to n (inclusive)

    .. seealso::

        - `SciPy trapezoid integration <https://docs.scipy.org/doc/scipy/reference/generated/scipy.integrate.cumulative_trapezoid.html>`_
          Documentation for the integration function used internally.

        - `SciPy Butterworth filter design <https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.butter.html>`_
          Documentation for the butterworth filter design function used in
          preprocessing.

        - `SciPy Tukey window <https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.windows.tukey.html>`_
          Documentation for the Tukey window function used in preprocessing.
    """
    return [
        integ
        for _, integ in zip(
            range(n + 1),
            iter_integrals(
                df,
                zero=zero,
                highpass_cutoff=highpass_cutoff,
                tukey_percent=tukey_percent,
            ),
        )
    ]



from __future__ import annotations

import typing
from typing import Any, Dict, List, Callable, Optional, Iterable

from dataclasses import dataclass
from functools import partial
import warnings
import os

import numpy as np
import pandas as pd

import endaq.ide
from endaq.calc import stats as calc_stats
from endaq.calc import psd as calc_psd

from endaq.batch import analyzer


def _make_meta(dataset):
    """Generate a pandas object containing metadata for the given recording."""
    serial_no = dataset.recorderInfo["RecorderSerial"]
    start_time = np.datetime64(dataset.sessions[0].utcStartTime, "s") + np.timedelta64(
        dataset.sessions[0].firstTime, "us"
    )

    return pd.Series(
        [serial_no, start_time],
        index=["serial number", "start time"],
        name=dataset.filename,
    )


def _make_psd(ch_data_cache: analyzer.CalcCache, fstart=None, bins_per_octave=None):
    """
    Format the PSD of the main accelerometer channel into a pandas object.

    The PSD is scaled to units of g^2/Hz (g := gravity = 9.80665 meters per
    square second).
    """
    df_psd = ch_data_cache._PSDData
    if df_psd.size == 0:
        return None

    if bins_per_octave is not None:
        df_psd = calc_psd.to_octave(
            df_psd,
            fstart=(fstart or 1),
            octave_bins=bins_per_octave,
            agg=np.mean,
        )

    df_psd["Resultant"] = np.sum(df_psd.to_numpy(), axis=1)
    df_psd = df_psd * analyzer.MPS2_TO_G ** 2  # (m/s^2)^2/Hz -> g^2/Hz

    return df_psd.stack(level="axis").reorder_levels(["axis", "frequency (Hz)"])


def _make_pvss(ch_data_cache: analyzer.CalcCache):
    """
    Format the PVSS of the main accelerometer channel into a pandas object.

    The PVSS is scaled to units of mm/sec.
    """
    df_pvss = ch_data_cache._PVSSData
    if df_pvss.size == 0:
        return None

    df_pvss["Resultant"] = ch_data_cache._PVSSResultantData
    df_pvss = df_pvss * analyzer.MPS_TO_MMPS

    return df_pvss.stack(level="axis").reorder_levels(["axis", "frequency (Hz)"])


def _make_halfsine_pvss_envelope(ch_data_cache, *args, **kwargs):
    df_pvss = ch_data_cache._PVSSData.copy()
    df_pvss["Resultant"] = ch_data_cache._PVSSResultantData
    df_pvss = df_pvss * analyzer.MPS_TO_MMPS
    if df_pvss.size == 0:
        return None

    return (
        endaq.calc.shock.enveloping_half_sine(df_pvss)
        .to_time_series(*args, **kwargs)
        .stack(level="axis")
        .reorder_levels(["axis", "timestamp"])
    )


def _make_metrics(
    ch_data_cache: analyzer.CalcCache,
    include: Iterable[str] = [],
    exclude: Iterable[str] = [],
):
    """
    Format the channel metrics of a recording into a pandas object.

    The following units listed by type are used for the metrics:
    - acceleration - g (gravity = 9.80665 meters per square second)
    - velocity - millimeters per second
    - displacement - millimeters
    - rotation speed - degrees per second
    - GPS position - degrees latitude/longitude
    - GPS speed - km/h (kilometers per hour)
    - audio - unitless
    - temperature - degrees Celsius
    - pressure - kiloPascals
    """
    if include and exclude:
        raise ValueError("parameters `include` and `exclude` are mutually-exclusive")

    DEFAULT_EXCLUDE = [x.casefold() for x in ["RMS Sound Pressure"]]
    VALID_METRICS = {
        k.casefold(): v
        for (k, v) in [
            ("RMS Acceleration", "accRMSFull"),
            ("RMS Velocity", "velRMSFull"),
            ("RMS Displacement", "disRMSFull"),
            ("Peak Absolute Acceleration", "accPeakFull"),
            ("Peak Pseudo Velocity Shock Spectrum", "pseudoVelPeakFull"),
            ("GPS Position", "gpsLocFull"),
            ("GPS Speed", "gpsSpeedFull"),
            ("RMS Angular Velocity", "gyroRMSFull"),
            ("RMS Sound Pressure", "micRMSFull"),
            ("Average Sound Pressure Level", "micDecibelsFull"),
            ("Average Temperature", "tempFull"),
            ("Average Pressure", "pressFull"),
            ("Average Relative Humidity", "humidFull"),
        ]
    }
    assert sorted(VALID_METRICS.values()) == sorted(
        attr
        for attr in dir(analyzer.CalcCache)
        if not attr.startswith("_") and attr.endswith("Full")
    )

    include = [x.casefold() for x in include]
    exclude = [x.casefold() for x in exclude]

    invalid_metrics = set(include or exclude) - set(VALID_METRICS)
    if invalid_metrics:
        raise ValueError(f"invalid metrics {list(invalid_metrics)}")

    metric_names: Iterable[str]
    if include:
        metric_names = include
    elif exclude:
        exclude = set(exclude)
        metric_names = (x for x in VALID_METRICS if x not in exclude)
    else:
        metric_names = (x for x in VALID_METRICS if x not in DEFAULT_EXCLUDE)
    metric_attrs = [VALID_METRICS[name] for name in metric_names]

    df = pd.concat(
        [getattr(ch_data_cache, attr) for attr in metric_attrs],
        axis="columns",
    )

    # Format data into desired shape
    df.columns.name = "calculation"
    series = df.stack().reorder_levels(["calculation", "axis"])

    return series


def _make_peak_windows(ch_data_cache: analyzer.CalcCache, margin_len):
    """
    Store windows of the main accelerometer channel about its peaks in a pandas
    object.

    The acceleration is scaled to units of g (gravity = 9.80665 meters per
    square second).
    """
    df_accel = ch_data_cache._accelerationData.copy()
    df_accel["Resultant"] = ch_data_cache._accelerationResultantData
    df_accel = analyzer.MPS2_TO_G * df_accel
    if df_accel.size == 0:
        return None

    dt = endaq.calc.sample_spacing(df_accel)

    data_noidx = df_accel.reset_index(drop=True)
    peak_indices = data_noidx.abs().idxmax(axis="rows")
    aligned_peak_data = pd.concat(
        [
            pd.Series(
                df_accel[col].to_numpy(),
                index=(data_noidx.index - peak_indices[col]),
                name=col,
            )
            for col in df_accel.columns
        ],
        axis="columns",
    )
    aligned_peak_data = aligned_peak_data.loc[-margin_len : margin_len + 1]
    aligned_peak_data.index = pd.Series(
        aligned_peak_data.index * dt, name="peak offset"
    )
    aligned_peak_data.columns = pd.MultiIndex.from_frame(
        pd.DataFrame(
            {
                "axis": aligned_peak_data.columns.values,
                "peak time": df_accel.index[peak_indices],
            }
        )
    )

    # Format results
    result = (
        aligned_peak_data.stack()
        .stack()
        .reorder_levels(["axis", "peak time", "peak offset"])
    )

    return result


def _make_vc_curves(ch_data_cache: analyzer.CalcCache):
    """
    Format the VC curves of the main accelerometer channel into a pandas object.
    """
    df_vc = ch_data_cache._VCCurveData * analyzer.MPS_TO_UMPS  # (m/s) -> (μm/s)
    df_vc["Resultant"] = calc_stats.L2_norm(df_vc.to_numpy(), axis=1)
    if df_vc.size == 0:
        return None

    return df_vc.stack(level="axis").reorder_levels(["axis", "frequency (Hz)"])


class GetDataBuilder:
    """
    The main interface for calculations in ``endaq.batch``.

    This object has two types of functions:

    - *configuration functions* - these determine what calculations will be
      performed on IDE recordings, and pass in any requisite parameters for said
      calculations. This includes the following functions:

      .. hlist::

          - :py:meth:`add_psd`
          - :py:meth:`add_pvss`
          - :py:meth:`add_pvss_halfsine_envelope`
          - :py:meth:`add_metrics`
          - :py:meth:`add_peaks`
          - :py:meth:`add_vc_curves`

    - *execution functions* - these functions take recording files as parameters,
      perform the configured calculations on the data therein, and return the
      calculated data as a :py:class:`.OutputStruct` object that wraps pandas
      objects.

      This includes the functions :py:meth:`_get_data` &
      :py:meth:`aggregate_data`, which operates on one & multiple file(s),
      respectively.

    A typical use case will look something like this:

    .. code-block:: python

        filenames = [...]

        calc_output = (
            GetDataBuilder(accel_highpass_cutoff=1)
            .add_psd(freq_bin_width=1)
            .add_pvss(init_freq=1, bins_per_octave=12)
            .add_pvss_halfsine_envelope()
            .add_metrics()
            .add_peaks(margin_len=100)
            .add_vc_curves(init_freq=1, bins_per_octave=3)
            .aggregate_data(filenames)
        )
        file_data = calc_output.dataframes
    """

    def __init__(
        self,
        *,
        preferred_chs=[],
        accel_highpass_cutoff,
        accel_start_time=None,
        accel_end_time=None,
        accel_start_margin=None,
        accel_end_margin=None,
        accel_integral_tukey_percent=0,
        accel_integral_zero="start",
    ):
        """
        :param preferred_chs: a sequence of channels; each channel listed is
            prioritized over others of the same type of physical measurement
            (e.g., acceleration, temperature, pressure, etc.)
        :param accel_highpass_cutoff: the cutoff frequency used when
            pre-filtering acceleration data
        :param accel_start_time: the relative timestamp before which to reject
            recording data; cannot be used in conjunction with
            `accel_start_margin`
        :param accel_end_time: the relative timestamp after which to reject
            recording data; cannot be used in conjunction with
            `accel_end_margin`
        :param accel_start_margin: the number of samples before which to reject
            recording data; cannot be used in conjunction with
            `accel_start_time`
        :param accel_end_margin: the number of samples after which to reject
            recording data; cannot be used in conjunction with
            `accel_end_time`
        :param accel_integral_tukey_percent: the alpha parameter of a Tukey
            window applied to the acceleration before integrating into
            velocity & displacement; see the `tukey_percent` parameter in
            :py:func:`endaq.calc.integrate.integrals` for details
        :param accel_integral_zero: the output quantity driven to zero when
            integrating the acceleration into velocity & displacement; see the
            `zero` parameter in :py:func:`endaq.calc.integrate.integrals` for
            details
        """
        if accel_start_time is not None and accel_start_margin is not None:
            raise ValueError(
                "only one of `accel_start_time` and `accel_start_margin` may be set at once"
            )
        if accel_end_time is not None and accel_end_margin is not None:
            raise ValueError(
                "only one of `accel_end_time` and `accel_end_margin` may be set at once"
            )

        self._metrics_queue: List[Tuple[str, Callable[[analyzer.CalcCache], Any]]] = []

        self._ch_data_cache_kwargs = dict(
            accel_highpass_cutoff=accel_highpass_cutoff,
            accel_start_time=accel_start_time,
            accel_end_time=accel_end_time,
            accel_start_margin=accel_start_margin,
            accel_end_margin=accel_end_margin,
            accel_integral_tukey_percent=accel_integral_tukey_percent,
            accel_integral_zero=accel_integral_zero,
        )
        self._preferred_chs = preferred_chs

        # Even unused parameters MUST be set; used to instantiate `CalcCache` in `_get_data`
        self._psd_freq_bin_width = None
        self._psd_freq_bin_width_oct = None
        self._psd_window = None
        self._pvss_init_freq = None
        self._pvss_bins_per_octave = None
        self._vc_init_freq = None
        self._vc_bins_per_octave = None

    def add_psd(
        self,
        freq_bin_width: Optional[float] = None,
        freq_start_octave: Optional[float] = None,
        bins_per_octave: Optional[float] = None,
        window: Optional[str] = None,
    ):
        """
        Add the acceleration PSD to the calculation queue.

        *calculation output units*: :math:`\\frac{\\text{G}^2}{\\text{Hz}}`,
        where `G` is the acceleration of gravity :math:`\\left( 1 \\text{G}
        \\approx 9.80665 \\frac{ \\text{m} }{ \\text{sec}^2 } \\right)`

        :param freq_bin_width: the desired spacing between adjacent PSD samples;
            a default is provided only if `bins_per_octave` is used, otherwise
            this parameter is required
        :param freq_start_octave: the first frequency to use in octave-spacing;
            this is only used if `bins_per_octave` is set
        :param bins_per_octave: the number of frequency bins per octave in a
            log-spaced PSD; if not set, the PSD will be linearly-spaced as
            specified by `freq_bin_width`
        :param window: the window type used in the PSD calculation; see the
            documentation for ``scipy.signal.welch`` for details
        """
        if all(i is None for i in (freq_bin_width, bins_per_octave)):
            raise ValueError(
                "must at least provide parameters for one of linear and log-spaced modes"
            )
        if freq_bin_width is None:
            if freq_start_octave is None:
                freq_start_octave = 1

            freq_bin_width = endaq.calc.psd._aligned_bin_width(
                freq_start_octave, bins_per_octave
            )
            self._psd_freq_bin_width_oct = min(
                freq_bin_width, self._psd_freq_bin_width_oct or float("inf")
            )
        else:
            self._psd_freq_bin_width = freq_bin_width

        if not (self._psd_window is None or self._psd_window == window):
            raise ValueError(
                "inconsistent PSD windows provided:"
                f" first {self._psd_window}, then {window}"
            )
        self._psd_window = window

        self._metrics_queue.append(
            (
                "psd",
                partial(
                    _make_psd,
                    fstart=freq_start_octave,
                    bins_per_octave=bins_per_octave,
                ),
            )
        )

        return self

    def add_pvss(self, init_freq: float = 1.0, bins_per_octave: float = 3.0):
        """
        Add the acceleration PVSS (Pseudo Velocity Shock Spectrum) to the
        calculation queue.

        *calculation output units*: :math:`\\frac{\\text{mm}}{\\text{sec}}`

        :param init_freq: the first frequency sample in the spectrum
        :param bins_per_octave: the number of samples per frequency octave
        """
        if any(name == "pvss" for (name, _) in self._metrics_queue):
            raise RuntimeError('cannot call "add_pvss" twice')

        self._metrics_queue.append(("pvss", _make_pvss))
        self._pvss_init_freq = init_freq
        self._pvss_bins_per_octave = bins_per_octave

        return self

    def add_pvss_halfsine_envelope(
        self,
        tstart: Optional[float] = None,
        tstop: Optional[float] = None,
        dt: Optional[float] = None,
        tpulse: Optional[float] = None,
    ):
        """
        Add the half-sine envelope for the acceleration's PVSS (Pseudo Velocity
        Shock Spectrum) to the calculation queue.

        *calculation output units*: :math:`\\frac{\\text{mm}}{\\text{sec}}`
        """
        self._metrics_queue.append(
            (
                "halfsine",
                partial(
                    _make_halfsine_pvss_envelope,
                    tstart=tstart,
                    tstop=tstop,
                    dt=dt,
                    tpulse=tpulse,
                ),
            )
        )

        return self

    def add_metrics(self, include: List[str] = [], exclude: List[str] = []):
        """
        Add broad channel metrics to the calculation queue.

        The output units for each metric are listed below:

        .. hlist::

            - `RMS Acceleration`: :math:`\\text{G}`
            - `RMS Velocity`: :math:`\\frac{\\text{mm}}{\\text{sec}}`
            - `RMS Displacement`: :math:`\\text{mm}`
            - `Peak Absolute Acceleration`: :math:`\\text{G}`
            - `Peak Pseudo Velocity Shock Spectrum`: :math:`\\frac{\\text{mm}}{\\text{sec}}`
            - `GPS Position`: :math:`\\text{degrees}`
            - `GPS Speed`: :math:`\\frac{\\text{km}}{\\text{hr}}`
            - `RMS Angular Velocity`: :math:`\\frac{\\text{degrees}}{\\text{sec}}`
            - `RMS Microphone`: :math:`\\text{Pascals}`
            - `Average Temperature`: :math:`{}^{\\circ} \\text{C}`
            - `Average Pressure`: :math:`\\text{Pascals}`
            - `Average Relative Humidity`: :math:`\\text{%}`

        where `G` is the acceleration of gravity :math:`\\left( 1 \\text{G}
        \\approx 9.80665 \\frac{ \\text{m} }{ \\text{sec}^2 } \\right)`

        """
        self._metrics_queue.append(
            ("metrics", partial(_make_metrics, include=include, exclude=exclude))
        )

        # no PSD metrics -> no need to provide PSD params

        # Need to provide default PVSS metrics
        if self._pvss_init_freq is None:
            self._pvss_init_freq = 1
            self._pvss_bins_per_octave = 12

        return self

    def add_peaks(self, margin_len: int = 1000):
        """
        Add windows about the acceleration's peak value to the calculation
        queue.

        *calculation output units*: :math:`\\text{G}`, where `G` is the
        acceleration of gravity :math:`\\left( 1 \\text{G} \\approx 9.80665
        \\frac{ \\text{m} }{ \\text{sec}^2 } \\right)`

        :param margin_len: the number of samples on each side of a peak to
            include in the windows
        """
        self._metrics_queue.append(
            (
                "peaks",
                partial(
                    _make_peak_windows,
                    margin_len=margin_len,
                ),
            )
        )

        return self

    def add_vc_curves(self, init_freq: float = 1.0, bins_per_octave: float = 3.0):
        """
        Add Vibration Criteria (VC) Curves to the calculation queue.

        *calculation output units*: :math:`\\frac{\\text{μm}}{\\text{sec}}`

        :param init_freq: the first frequency
        :param bins_per_octave:  the number of samples per frequency octave
        """
        self._metrics_queue.append(("vc_curves", _make_vc_curves))

        if "psd" not in self._metrics_queue:
            self._psd_freq_bin_width_oct = min(
                0.2,  # TODO: use `endaq.calc.psd._aligned_bin_width`
                self._psd_freq_bin_width_oct or float("inf"),
            )
        self._vc_init_freq = init_freq
        self._vc_bins_per_octave = bins_per_octave

        return self

    def _make_calc_params(self) -> analyzer.CalcParams:
        return analyzer.CalcParams(
            **self._ch_data_cache_kwargs,
            psd_window=self._psd_window or "hanning",
            psd_freq_bin_width=self._psd_freq_bin_width or self._psd_freq_bin_width_oct,
            pvss_init_freq=self._pvss_init_freq,
            pvss_bins_per_octave=self._pvss_bins_per_octave,
            vc_init_freq=self._vc_init_freq,
            vc_bins_per_octave=self._vc_bins_per_octave,
        )

    def _get_data(self, filename):
        """
        Calculate data from a single recording into a pandas object.

        Used internally by `aggregate_data`.
        """
        print(f"processing {filename}...")

        data = []
        with endaq.ide.get_doc(filename) as ds:
            ch_data_cache = analyzer.CalcCache.from_ide(
                ds,
                self._make_calc_params(),
                preferred_chs=self._preferred_chs,
            )

            data.append(("meta", _make_meta(ds)))

            for output_type, func in self._metrics_queue:
                data.append((output_type, func(ch_data_cache)))

        return data

    def aggregate_data(self, filenames) -> Optional[OutputStruct]:
        """
        Compile configured data from the given files into a dataframe.

        :param filenames: a sequence of paths of recording files to process
        """
        if len(filenames) == 0:
            return None

        http_files, local_files = [], []
        for file in filenames:
            path_formatted, mode = endaq.ide.files.normalized_path(file)

            if mode == "url":
                http_files.append(file)
            else:  # mode == "local"
                local_files.append(path_formatted)

        if len(local_files) == 0:
            root_path = ""
        elif len(local_files) == 1:
            # Common path will take the one file's whole path as the "root path"
            # -> remove the basename from this path
            root_path = os.path.dirname(local_files[0])
        else:
            root_path = os.path.commonpath(local_files)

        files = http_files + local_files
        display_names = http_files + [
            os.path.relpath(name, start=root_path) for name in local_files
        ]

        series_lists = zip(*([d for (_k, d) in self._get_data(file)] for file in files))

        print("aggregating data...")
        meta, *dfs = (
            pd.concat(
                series_list,
                keys=display_names,
                names=["filename"]
                + next(s for s in series_list if s is not None).index.names,
            )
            if series_list and any(s is not None for s in series_list)
            else None
            for series_list in series_lists
        )

        meta = meta.unstack(level=1)
        meta.attrs["rootpath"] = root_path

        def reformat(series):
            if series is None:
                return None

            df = series.to_frame().T.melt()
            df["serial number"] = meta.loc[df["filename"], "serial number"].reset_index(
                drop=True
            )
            df["start time"] = meta.loc[df["filename"], "start time"].reset_index(
                drop=True
            )

            return df

        dfs = [("meta", meta)] + [
            (df_type, reformat(df))
            for ((df_type, _), df) in zip(self._metrics_queue, dfs)
        ]

        print("done!")

        return OutputStruct(dfs)


class OutputStruct:
    """
    A data wrapper class with methods for common export operations.

    Objects of this class are generated by :py:meth:`.GetDataBuilder.aggregate_data`.
    This class is not intended be instantiated manually.
    """

    def __init__(self, data: List[Tuple[str, pd.DataFrame]]):
        self.dataframes = data

    def to_csv_folder(self, folder_path):
        """
        Write data to a folder as CSV's.

        :param folder_path: the output directory path for .CSV files
        """
        os.makedirs(folder_path, exist_ok=True)

        for k, df in self.dataframes:
            path = os.path.join(folder_path, f"{k}.csv")
            df.to_csv(path, index=(k == "meta"))

    def to_html_plots(
        self,
        folder_path=None,
        show: bool = False,
        theme: typing.Literal[
            None, "endaq", "endaq_light", "endaq_arial", "endaq_light_arial"
        ] = "endaq",
    ):
        """
        Generate plots in HTML.

        :param folder_path: The output directory for saving .HTML
            plots. If `None` (default), plots are not saved.
        :param show: Whether to open plots after generation. Defaults to `False`.
        :param theme: The enDAQ plotly theme to use; see
            :py:func:`endaq.plot.utilities.set_theme` for details on the
            supported options. Defaults to `"endaq"`. If `None`, the default
            Plotly theme is used.
        """
        if not any((folder_path, show)):
            return

        import plotly.express as px

        if theme is not None:
            from endaq.plot.utilities import set_theme

            set_theme(theme)

        if folder_path:
            os.makedirs(folder_path, exist_ok=True)

        for k, df in self.dataframes:
            if k == "meta":
                continue
            if k == "psd":
                fig = px.line(
                    df,
                    x="frequency (Hz)",
                    y="value",
                    color="filename",
                    line_dash="axis",
                )

                fig.update_xaxes(type="log", title_text="frequency (Hz)")
                fig.update_yaxes(type="log", title_text="Acceleration (g^2/Hz)")
                fig.update_layout(title="Acceleration PSD")

            elif k == "pvss":
                fig = px.line(
                    df,
                    x="frequency (Hz)",
                    y="value",
                    color="filename",
                    line_dash="axis",
                )
                fig.update_xaxes(type="log", title_text="frequency (Hz)")
                fig.update_yaxes(type="log", title_text="Velocity (mm/s)")
                fig.update_layout(title="Pseudo Velocity Shock Spectrum (PVSS)")

            elif k == "halfsine":
                fig = px.line(
                    df,
                    x="timestamp",
                    y="value",
                    color="filename",
                    line_dash="axis",
                )
                fig.update_layout(title="PVSS Enveloping Half Sine Pulses")

            elif k == "metrics":
                warnings.warn("HTML plot for metrics not currently implemented")
                continue

            elif k == "peaks":
                fig = px.line(
                    df,
                    x=df["peak offset"],
                    y="value",
                    color="filename",
                    line_dash="axis",
                )
                fig.update_xaxes(title_text="time relative to peak (s)")
                fig.update_yaxes(title_text="Acceleration (g)")
                fig.update_layout(title="Window about Acceleration Peaks")

            elif k == "vc_curves":
                fig = px.line(
                    df,
                    x="frequency (Hz)",
                    y="value",
                    color="filename",
                    line_dash="axis",
                )

                fig.update_xaxes(type="log", title_text="frequency (Hz)")
                fig.update_yaxes(
                    type="log", title_text="1/3-Octave RMS Velocity (μm/s)"
                )
                fig.update_layout(title="Vibration Criteria (VC) Curves")

            else:
                raise RuntimeError(f"no configuration for plotting '{k}' data")

            if not folder_path and show:
                fig.show()
            else:
                fig.write_html(
                    file=os.path.join(folder_path, f"{k}.html"),
                    include_plotlyjs="directory",
                    full_html=True,
                    auto_open=show,
                )



import numpy as np


def as_quat_array(q, force_copy=False, qaxis=-1):
    """Formats a quaternion array."""
    q = (np.array if force_copy else np.asarray)(q)

    if not (-q.ndim <= qaxis < q.ndim):
        raise ValueError("invalid quaternion axis")

    if q.ndim == 0 or q.shape[qaxis] != 4:
        raise ValueError(
            "improper shape for quaternion data;"
            " specified axis should contain WXYZ coordinates"
        )

    return q


def quat_mul(q1, q2, qaxis=-1):
    """Multiply two quaternion arrays."""
    q1, q2 = as_quat_array(q1, qaxis=qaxis), as_quat_array(q2, qaxis=qaxis)
    q1, q2 = np.moveaxis(q1, qaxis, -1), np.moveaxis(q2, qaxis, -1)

    result = np.empty(np.broadcast(q1, q2).shape, dtype=np.result_type(q1, q2))

    result[..., 0] = np.sum(q1 * q2 * np.array([1, -1, -1, -1]), axis=-1)
    result[..., 1] = np.sum(
        q1 * q2[..., [1, 0, 3, 2]] * np.array([1, 1, 1, -1]), axis=-1
    )
    result[..., 2] = np.sum(
        q1 * q2[..., [2, 3, 0, 1]] * np.array([1, -1, 1, 1]), axis=-1
    )
    result[..., 3] = np.sum(
        q1 * q2[..., [3, 2, 1, 0]] * np.array([1, 1, -1, 1]), axis=-1
    )

    return np.moveaxis(result, -1, qaxis)


def quat_conj(q, qaxis=-1):
    """Conjugate an array of quaternion."""
    q = as_quat_array(q, force_copy=True, qaxis=qaxis)

    q = np.moveaxis(q, qaxis, 0)
    q[1:] *= -1
    return np.moveaxis(q, 0, qaxis)


def quat_inv(q, qaxis=-1):
    """Invert an array of quaternions."""
    return quat_conj(q, qaxis=qaxis) / np.sum(q ** 2, axis=qaxis, keepdims=True)


def quat_div(q1, q2, qaxis=-1):
    """Divide two quaternion arrays."""
    return quat_mul(q1, quat_inv(q2, qaxis=qaxis), qaxis=qaxis)


def quat_to_angvel(q, *dt, qaxis=-1):
    """
    Calculate the angular velocity for an array of orientation quaternions.

    :param q: quaternion array; requires q.shape[-1] == 4
    :param dt: the time corresponing to each quaternion sample
    :return: the angular velocity
    """
    q = np.moveaxis(q, qaxis, -1)
    q_prime = np.gradient(q, *dt, axis=range(q.ndim - 1), edge_order=2)
    ang_vel = 2 * quat_div(q_prime, q)[..., 1:]

    return np.moveaxis(ang_vel, -1, qaxis)



"""
Some general-purpose IDE file manipulation funcions.
"""
import datetime
import string
from ebmlite import loadSchema

__all__ = ['parse_time', 'validate']


# ============================================================================
#
# ============================================================================

def validate(stream, from_pos=False, lookahead=25, percent=.5):
    """
    Determine if a stream contains IDE data.

    :param stream: A file-like stream (something that supports the methods
        `tell()` and `seek()`).
    :param from_pos: If `True`, validation of the stream will start at its
        current position. If `False` (default), the validation will start
        from its beginning.
    :param lookahead: The number of EBML elements to check.
    :param percent: The minumum percentage of EBML elements identified as
        being part of the IDE schema for validation. A small number of
        unknown elements may not indicate an invalid file; it may simply
        have been created using a different version of the schema.
    :return: `True` if validation passed, `False` if it failed.
    """
    # TODO: Make validation more thorough, test for corrupt files?

    orig_pos = stream.tell()
    if not from_pos:
        stream.seek(0)

    try:
        schema = loadSchema('mide_ide.xml')
        doc = schema.load(stream, headers=True)

        # Basic test: is it EBML data in the expected schema?
        known = 0
        for idx, el in enumerate(doc):
            if idx >= lookahead:
                break
            if el.id in schema.elements:
                known += 1
        if known < lookahead * percent:
            return False

        return True

    finally:
        stream.seek(orig_pos)


# ============================================================================
#
# ============================================================================

def parse_time(t, datetime_start=None):
    """ Convert a time in one of several user-friendly forms to microseconds
        (the native time units used in `idelib`). Valid types are:

        * `None`, `int`, or `float` (returns the same value)
        * `str` (formatted as a time, e.g., `MM:SS`, `HH:MM:SS`,
          `DDd HH:MM:SS`). More examples:

          * ``":01"`` or ``":1"`` or ``"1s"`` (1 second)
          * ``"22:11"`` (22 minutes, 11 seconds)
          * ``"3:22:11"`` (3 hours, 22 minutes, 11 seconds)
          * ``"1d 3:22:11"`` (3 hours, 22 minutes, 11 seconds)
        * `datetime.timedelta` or `pandas.Timedelta`
        * `datetime.datetime`

        :param t: The time value to convert.
        :param datetime_start: If `t` is a `datetime` object, the result will
            be relative to `datetime_start`. It will default to the start of
            the day portion of `t`. This has no effect on non-`datetime`
            values of `t` .
        :returns: The time in microseconds.
    """
    # TODO: Put this somewhere else? It will be useful elsewhere, and shouldn't
    #   be bound to the `pandas` requirement in this module.

    if t is None or isinstance(t, (int, float)):
        return t

    elif isinstance(t, str):
        if not t:
            return None
        orig = t
        t = t.strip().lower()
        for c in ":dhms":
            t = t.replace(c, ' ')
        if not all(c in string.digits + ' ' for c in t):
            raise ValueError(f"Bad time string for parse_time(): {orig!r}")

        micros = 0
        for part, mult in zip(reversed(t.split()), (1, 60, 3600, 86400)):
            if not part:
                continue
            part = part.strip(string.ascii_letters + string.punctuation + string.whitespace)
            micros += float(part) * mult
        return micros * 10**6

    elif isinstance(t, datetime.timedelta):
        return t.total_seconds() * 10**6

    elif isinstance(t, (datetime.time, datetime.datetime)):
        if datetime_start is None:
            # No starting time, assume midnight of same day.
            datetime_start = datetime.datetime(t.year, t.month, t.day)

        if isinstance(t, datetime.time):
            # just time: make datetime
            t = datetime.datetime.combine(datetime_start, t)

        if isinstance(t, datetime.datetime):
            # datetime: make timedelta
            return (t - datetime_start).total_seconds() * 10**6

    raise TypeError(f"Unsupported type for parse_time(): {type(t).__name__} ({t!r})")





import requests
from types import FunctionType
import plotly.io as pio
import json

from .core import EndaqCloud, ENV_PRODUCTION, ENV_STAGING, ENV_DEVELOP


__all__ = [
    'create_cloud_dashboard_output',
    'produce_dashboard_plots',
]


def create_cloud_dashboard_output(name_to_fig: dict) -> str:
    """
    A function which makes producing the json based string used to produce custom enDAQ Cloud report
    dashboards easy.

    :param name_to_fig: A dictionary mapping the desired names/titles of plots to a Plotly figure.
     The dictionary must have 4 elements, and the ordering DOES matter as it dictates the order
     the plots will be added to the dashboard.  This means the ordering of Python dictionaries which
     appears in Python 3.7 is crucial for using this!
    :return: The json based string which is to be given as the variable 'output' in enDAQ cloud
     custom dashboards
    """
    if not isinstance(name_to_fig, dict):
        raise TypeError(f"'name_to_fig' parameter must be a dictionary, but was given {type(name_to_fig)} instead")

    if len(name_to_fig) != 4:
        raise ValueError("The number of (key, value) pairs in 'name_to_fig' must be exactly 4, "
                         f"but {len(name_to_fig)} are given")

    return "[" + ", ".join([v.to_json()[:-1] + ', "title": "' + k + '"}' for k, v in name_to_fig.items()]) + "]"


def produce_dashboard_plots(dashboard_script_fn: FunctionType, api_key: str, max_num_files: int = 100,
                            environment: str = 'production', display_plots: bool = True) -> list:
    """
    A function used to simulate a run of a desired enDAQ Cloud custom report script without needing to use
    ``cloud.endaq.com``.

    :param dashboard_script_fn: A function accepting the parameters `files` and `file_download_url`, which has all
     the exact code that would be put into a enDAQ Cloud custom report script, followed by one final line:
     `return output`
    :param api_key: The enDAQ Cloud API key
    :param max_num_files: The maximum number of files to get data about.  Specifically, this is used to
     specify how many of the most recently uploaded IDE files in the cloud will have their info passed to
     your custom report script (through a list of json blobs, as parameter 'files')
    :param environment: The version of the enDAQ Cloud to communicate with, the options are 'production', 'staging',
     or 'develop'.  This should only be used internally at Mide
    :param display_plots: If the plots being produced should be displayed
    :return: A list of the 4 plotly figures produced
    """
    if environment == 'production':
        api_access_url = ENV_PRODUCTION
    elif environment == 'staging':
        api_access_url = ENV_STAGING
    elif environment == 'develop':
        api_access_url = ENV_DEVELOP
    else:
        raise ValueError("Only 'production', 'staging', and 'develop' may be given for the 'environment' parameter, "
                         f" but {environment} was given instead.")

    parameters = {"x-api-key": api_key}

    cloud_obj = EndaqCloud(api_key, env=api_access_url)

    files = cloud_obj._get_files_json_response(limit=max_num_files)

    most_recent_file_id = files[-1]['id']

    file_download_url = requests.get(
        api_access_url + '/api/v1/files/download/' + most_recent_file_id,
        headers=parameters
    ).json()['url']

    output = dashboard_script_fn(files=files, file_download_url=file_download_url)

    figures = [pio.from_json(json.dumps(blob)) for blob in json.loads(output)]

    if display_plots:
        for fig in figures:
            fig.show()

    return figures
